# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.environ as pyo
from pyomo.core.expr.visitor import identify_variables
from pyomo.common.dependencies import (
    networkx_available,
    plotly_available,
    scipy_available,
)
from pyomo.common.collections import ComponentSet, ComponentMap
from pyomo.contrib.incidence_analysis.interface import (
    asl_available,
    IncidenceGraphInterface,
    get_structural_incidence_matrix,
    get_numeric_incidence_matrix,
    get_incidence_graph,
    get_bipartite_incidence_graph,
    extract_bipartite_subgraph,
)
from pyomo.contrib.incidence_analysis.matching import maximum_matching
from pyomo.contrib.incidence_analysis.triangularize import (
    map_coords_to_block_triangular_indices,
)
from pyomo.contrib.incidence_analysis.dulmage_mendelsohn import dulmage_mendelsohn
from pyomo.contrib.incidence_analysis.tests.models_for_testing import (
    make_gas_expansion_model,
    make_degenerate_solid_phase_model,
    make_dynamic_model,
)

if scipy_available:
    from pyomo.contrib.pynumero.interfaces.pyomo_nlp import PyomoNLP
if networkx_available:
    import networkx as nx
    from networkx.algorithms.bipartite.matrix import from_biadjacency_matrix

import pyomo.common.unittest as unittest


@unittest.skipUnless(networkx_available, "networkx is not available.")
@unittest.skipUnless(scipy_available, "scipy is not available.")
@unittest.skipUnless(asl_available, "pynumero PyomoNLP is not available")
class TestGasExpansionNumericIncidenceMatrix(unittest.TestCase):
    """
    This class tests the get_numeric_incidence_matrix function on
    the gas expansion model.
    """

    def test_incidence_matrix(self):
        N = 5
        model = make_gas_expansion_model(N)
        all_vars = list(model.component_data_objects(pyo.Var))
        all_cons = list(model.component_data_objects(pyo.Constraint))
        imat = get_numeric_incidence_matrix(all_vars, all_cons)
        n_var = 4 * (N + 1)
        n_con = 4 * N + 1
        self.assertEqual(imat.shape, (n_con, n_var))

        var_idx_map = ComponentMap((v, i) for i, v in enumerate(all_vars))
        con_idx_map = ComponentMap((c, i) for i, c in enumerate(all_cons))

        # Map constraints to the variables they contain.
        csr_map = ComponentMap()
        csr_map.update(
            (
                model.mbal[i],
                ComponentSet(
                    [model.F[i], model.F[i - 1], model.rho[i], model.rho[i - 1]]
                ),
            )
            for i in model.streams
            if i != model.streams.first()
        )
        csr_map.update(
            (
                model.ebal[i],
                ComponentSet(
                    [
                        model.F[i],
                        model.F[i - 1],
                        model.rho[i],
                        model.rho[i - 1],
                        model.T[i],
                        model.T[i - 1],
                    ]
                ),
            )
            for i in model.streams
            if i != model.streams.first()
        )
        csr_map.update(
            (
                model.expansion[i],
                ComponentSet(
                    [model.rho[i], model.rho[i - 1], model.P[i], model.P[i - 1]]
                ),
            )
            for i in model.streams
            if i != model.streams.first()
        )
        csr_map.update(
            (model.ideal_gas[i], ComponentSet([model.P[i], model.rho[i], model.T[i]]))
            for i in model.streams
        )

        # Map constraint and variable indices to the values of the derivatives
        # Note that the derivative values calculated here depend on the model's
        # canonical form.
        deriv_lookup = {}
        m = model  # for convenience
        for s in model.streams:
            # Ideal gas:
            i = con_idx_map[model.ideal_gas[s]]
            j = var_idx_map[model.P[s]]
            deriv_lookup[i, j] = 1.0

            j = var_idx_map[model.rho[s]]
            deriv_lookup[i, j] = -model.R.value * model.T[s].value

            j = var_idx_map[model.T[s]]
            deriv_lookup[i, j] = -model.R.value * model.rho[s].value

            if s != model.streams.first():
                # Expansion:
                i = con_idx_map[model.expansion[s]]
                j = var_idx_map[model.P[s]]
                deriv_lookup[i, j] = 1 / model.P[s - 1].value

                j = var_idx_map[model.P[s - 1]]
                deriv_lookup[i, j] = -model.P[s].value / model.P[s - 1] ** 2

                j = var_idx_map[model.rho[s]]
                deriv_lookup[i, j] = pyo.value(
                    -m.gamma * (m.rho[s] / m.rho[s - 1]) ** (m.gamma - 1) / m.rho[s - 1]
                )

                j = var_idx_map[model.rho[s - 1]]
                deriv_lookup[i, j] = pyo.value(
                    -m.gamma
                    * (m.rho[s] / m.rho[s - 1]) ** (m.gamma - 1)
                    * (-m.rho[s] / m.rho[s - 1] ** 2)
                )

                # Energy balance:
                i = con_idx_map[m.ebal[s]]
                j = var_idx_map[m.rho[s - 1]]
                deriv_lookup[i, j] = pyo.value(m.F[s - 1] * m.T[s - 1])

                j = var_idx_map[m.F[s - 1]]
                deriv_lookup[i, j] = pyo.value(m.rho[s - 1] * m.T[s - 1])

                j = var_idx_map[m.T[s - 1]]
                deriv_lookup[i, j] = pyo.value(m.F[s - 1] * m.rho[s - 1])

                j = var_idx_map[m.rho[s]]
                deriv_lookup[i, j] = pyo.value(-m.F[s] * m.T[s])

                j = var_idx_map[m.F[s]]
                deriv_lookup[i, j] = pyo.value(-m.rho[s] * m.T[s])

                j = var_idx_map[m.T[s]]
                deriv_lookup[i, j] = pyo.value(-m.F[s] * m.rho[s])

                # Mass balance:
                i = con_idx_map[m.mbal[s]]
                j = var_idx_map[m.rho[s - 1]]
                deriv_lookup[i, j] = pyo.value(m.F[s - 1])

                j = var_idx_map[m.F[s - 1]]
                deriv_lookup[i, j] = pyo.value(m.rho[s - 1])

                j = var_idx_map[m.rho[s]]
                deriv_lookup[i, j] = pyo.value(-m.F[s])

                j = var_idx_map[m.F[s]]
                deriv_lookup[i, j] = pyo.value(-m.rho[s])

        # Want to test that the columns have the rows we expect.
        i = model.streams.first()
        for i, j, e in zip(imat.row, imat.col, imat.data):
            con = all_cons[i]
            var = all_vars[j]
            self.assertIn(var, csr_map[con])
            csr_map[con].remove(var)
            self.assertAlmostEqual(pyo.value(deriv_lookup[i, j]), pyo.value(e), 8)
        # And no additional rows
        for con in csr_map:
            self.assertEqual(len(csr_map[con]), 0)

    #
    # The following tests were copied from the
    # TestGasExpansionModelInterfaceClass test cases, and modified
    # to use the data format returned by get_numeric_incidence_matrix.
    #
    def test_imperfect_matching(self):
        model = make_gas_expansion_model()
        all_vars = list(model.component_data_objects(pyo.Var))
        all_cons = list(model.component_data_objects(pyo.Constraint))
        imat = get_numeric_incidence_matrix(all_vars, all_cons)

        n_eqn = len(all_cons)
        matching = maximum_matching(imat)
        values = set(matching.values())
        self.assertEqual(len(matching), n_eqn)
        self.assertEqual(len(values), n_eqn)

    def test_perfect_matching(self):
        model = make_gas_expansion_model()

        # These are the variables and constraints of the square,
        # nonsingular subsystem
        variables = []
        variables.extend(model.P.values())
        variables.extend(
            model.T[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.rho[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.F[i] for i in model.streams if i != model.streams.first()
        )

        constraints = list(model.component_data_objects(pyo.Constraint))

        imat = get_numeric_incidence_matrix(variables, constraints)
        con_idx_map = ComponentMap((c, i) for i, c in enumerate(constraints))

        n_var = len(variables)
        matching = maximum_matching(imat)
        matching = ComponentMap(
            (c, variables[matching[con_idx_map[c]]]) for c in constraints
        )
        values = ComponentSet(matching.values())
        self.assertEqual(len(matching), n_var)
        self.assertEqual(len(values), n_var)

        # The subset of variables and equations we have identified
        # do not have a unique perfect matching. But we at least know
        # this much.
        self.assertIs(matching[model.ideal_gas[0]], model.P[0])

    def test_triangularize(self):
        N = 5
        model = make_gas_expansion_model(N)

        # These are the variables and constraints of the square,
        # nonsingular subsystem
        variables = []
        variables.extend(model.P.values())
        variables.extend(
            model.T[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.rho[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.F[i] for i in model.streams if i != model.streams.first()
        )

        constraints = list(model.component_data_objects(pyo.Constraint))

        imat = get_numeric_incidence_matrix(variables, constraints)
        con_idx_map = ComponentMap((c, i) for i, c in enumerate(constraints))
        var_idx_map = ComponentMap((v, i) for i, v in enumerate(variables))

        row_block_map, col_block_map = map_coords_to_block_triangular_indices(imat)
        var_block_map = ComponentMap(
            (v, col_block_map[var_idx_map[v]]) for v in variables
        )
        con_block_map = ComponentMap(
            (c, row_block_map[con_idx_map[c]]) for c in constraints
        )

        var_values = set(var_block_map.values())
        con_values = set(con_block_map.values())
        self.assertEqual(len(var_values), N + 1)
        self.assertEqual(len(con_values), N + 1)

        self.assertEqual(var_block_map[model.P[0]], 0)

        for i in model.streams:
            if i != model.streams.first():
                self.assertEqual(var_block_map[model.rho[i]], i)
                self.assertEqual(var_block_map[model.T[i]], i)
                self.assertEqual(var_block_map[model.P[i]], i)
                self.assertEqual(var_block_map[model.F[i]], i)

                self.assertEqual(con_block_map[model.ideal_gas[i]], i)
                self.assertEqual(con_block_map[model.expansion[i]], i)
                self.assertEqual(con_block_map[model.mbal[i]], i)
                self.assertEqual(con_block_map[model.ebal[i]], i)


@unittest.skipUnless(networkx_available, "networkx is not available.")
@unittest.skipUnless(scipy_available, "scipy is not available.")
class TestGasExpansionStructuralIncidenceMatrix(unittest.TestCase):
    """
    This class tests the get_structural_incidence_matrix function
    on the gas expansion model.
    """

    def test_incidence_matrix(self):
        N = 5
        model = make_gas_expansion_model(N)
        all_vars = list(model.component_data_objects(pyo.Var))
        all_cons = list(model.component_data_objects(pyo.Constraint))
        imat = get_structural_incidence_matrix(all_vars, all_cons)
        n_var = 4 * (N + 1)
        n_con = 4 * N + 1
        self.assertEqual(imat.shape, (n_con, n_var))

        var_idx_map = ComponentMap((v, i) for i, v in enumerate(all_vars))
        con_idx_map = ComponentMap((c, i) for i, c in enumerate(all_cons))

        # Map constraints to the variables they contain.
        csr_map = ComponentMap()
        csr_map.update(
            (
                model.mbal[i],
                ComponentSet(
                    [model.F[i], model.F[i - 1], model.rho[i], model.rho[i - 1]]
                ),
            )
            for i in model.streams
            if i != model.streams.first()
        )
        csr_map.update(
            (
                model.ebal[i],
                ComponentSet(
                    [
                        model.F[i],
                        model.F[i - 1],
                        model.rho[i],
                        model.rho[i - 1],
                        model.T[i],
                        model.T[i - 1],
                    ]
                ),
            )
            for i in model.streams
            if i != model.streams.first()
        )
        csr_map.update(
            (
                model.expansion[i],
                ComponentSet(
                    [model.rho[i], model.rho[i - 1], model.P[i], model.P[i - 1]]
                ),
            )
            for i in model.streams
            if i != model.streams.first()
        )
        csr_map.update(
            (model.ideal_gas[i], ComponentSet([model.P[i], model.rho[i], model.T[i]]))
            for i in model.streams
        )

        # Want to test that the columns have the rows we expect.
        i = model.streams.first()
        for i, j, e in zip(imat.row, imat.col, imat.data):
            con = all_cons[i]
            var = all_vars[j]
            self.assertIn(var, csr_map[con])
            csr_map[con].remove(var)
            self.assertEqual(e, 1.0)
        # And no additional rows
        for con in csr_map:
            self.assertEqual(len(csr_map[con]), 0)

    #
    # The following tests were copied from the
    # TestGasExpansionModelInterfaceClass test cases, and modified
    # to use the data format returned by get_structural_incidence_matrix.
    #
    def test_imperfect_matching(self):
        model = make_gas_expansion_model()
        all_vars = list(model.component_data_objects(pyo.Var))
        all_cons = list(model.component_data_objects(pyo.Constraint))
        imat = get_structural_incidence_matrix(all_vars, all_cons)

        n_eqn = len(all_cons)
        matching = maximum_matching(imat)
        values = set(matching.values())
        self.assertEqual(len(matching), n_eqn)
        self.assertEqual(len(values), n_eqn)

    def test_perfect_matching(self):
        model = make_gas_expansion_model()

        # These are the variables and constraints of the square,
        # nonsingular subsystem
        variables = []
        variables.extend(model.P.values())
        variables.extend(
            model.T[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.rho[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.F[i] for i in model.streams if i != model.streams.first()
        )

        constraints = list(model.component_data_objects(pyo.Constraint))

        imat = get_structural_incidence_matrix(variables, constraints)
        con_idx_map = ComponentMap((c, i) for i, c in enumerate(constraints))

        n_var = len(variables)
        matching = maximum_matching(imat)
        matching = ComponentMap(
            (c, variables[matching[con_idx_map[c]]]) for c in constraints
        )
        values = ComponentSet(matching.values())
        self.assertEqual(len(matching), n_var)
        self.assertEqual(len(values), n_var)

        # The subset of variables and equations we have identified
        # do not have a unique perfect matching. But we at least know
        # this much.
        self.assertIs(matching[model.ideal_gas[0]], model.P[0])

    def test_triangularize(self):
        N = 5
        model = make_gas_expansion_model(N)

        # These are the variables and constraints of the square,
        # nonsingular subsystem
        variables = []
        variables.extend(model.P.values())
        variables.extend(
            model.T[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.rho[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.F[i] for i in model.streams if i != model.streams.first()
        )

        constraints = list(model.component_data_objects(pyo.Constraint))

        imat = get_structural_incidence_matrix(variables, constraints)
        con_idx_map = ComponentMap((c, i) for i, c in enumerate(constraints))
        var_idx_map = ComponentMap((v, i) for i, v in enumerate(variables))

        row_block_map, col_block_map = map_coords_to_block_triangular_indices(imat)
        var_block_map = ComponentMap(
            (v, col_block_map[var_idx_map[v]]) for v in variables
        )
        con_block_map = ComponentMap(
            (c, row_block_map[con_idx_map[c]]) for c in constraints
        )

        var_values = set(var_block_map.values())
        con_values = set(con_block_map.values())
        self.assertEqual(len(var_values), N + 1)
        self.assertEqual(len(con_values), N + 1)

        self.assertEqual(var_block_map[model.P[0]], 0)

        for i in model.streams:
            if i != model.streams.first():
                self.assertEqual(var_block_map[model.rho[i]], i)
                self.assertEqual(var_block_map[model.T[i]], i)
                self.assertEqual(var_block_map[model.P[i]], i)
                self.assertEqual(var_block_map[model.F[i]], i)

                self.assertEqual(con_block_map[model.ideal_gas[i]], i)
                self.assertEqual(con_block_map[model.expansion[i]], i)
                self.assertEqual(con_block_map[model.mbal[i]], i)
                self.assertEqual(con_block_map[model.ebal[i]], i)


@unittest.skipUnless(networkx_available, "networkx is not available.")
@unittest.skipUnless(scipy_available, "scipy is not available.")
@unittest.skipUnless(asl_available, "pynumero PyomoNLP is not available")
class TestGasExpansionModelInterfaceClassNumeric(unittest.TestCase):
    # In these tests, we pass the interface a PyomoNLP and cache
    # its Jacobian.
    def test_imperfect_matching(self):
        model = make_gas_expansion_model()
        model.obj = pyo.Objective(expr=0)
        nlp = PyomoNLP(model)
        igraph = IncidenceGraphInterface(nlp)

        n_eqn = len(list(model.component_data_objects(pyo.Constraint)))
        matching = igraph.maximum_matching()
        values = ComponentSet(matching.values())
        self.assertEqual(len(matching), n_eqn)
        self.assertEqual(len(values), n_eqn)

    def test_perfect_matching(self):
        model = make_gas_expansion_model()
        model.obj = pyo.Objective(expr=0)
        nlp = PyomoNLP(model)
        igraph = IncidenceGraphInterface(nlp)

        # These are the variables and constraints of the square,
        # nonsingular subsystem
        variables = []
        variables.extend(model.P.values())
        variables.extend(
            model.T[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.rho[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.F[i] for i in model.streams if i != model.streams.first()
        )

        constraints = list(model.component_data_objects(pyo.Constraint))

        n_var = len(variables)
        matching = igraph.maximum_matching(variables, constraints)
        values = ComponentSet(matching.values())
        self.assertEqual(len(matching), n_var)
        self.assertEqual(len(values), n_var)

        # The subset of variables and equations we have identified
        # do not have a unique perfect matching. But we at least know
        # this much.
        self.assertIs(matching[model.ideal_gas[0]], model.P[0])

    def test_triangularize(self):
        N = 5
        model = make_gas_expansion_model(N)
        model.obj = pyo.Objective(expr=0)
        nlp = PyomoNLP(model)
        igraph = IncidenceGraphInterface(nlp)

        # These are the variables and constraints of the square,
        # nonsingular subsystem
        variables = []
        variables.extend(model.P.values())
        variables.extend(
            model.T[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.rho[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.F[i] for i in model.streams if i != model.streams.first()
        )

        constraints = list(model.component_data_objects(pyo.Constraint))

        var_blocks, con_blocks = igraph.block_triangularize(variables, constraints)
        partition = [
            list(zip(vblock, cblock)) for vblock, cblock in zip(var_blocks, con_blocks)
        ]
        self.assertEqual(len(partition), N + 1)

        for i in model.streams:
            variables = ComponentSet([var for var, _ in partition[i]])
            constraints = ComponentSet([con for _, con in partition[i]])
            if i == model.streams.first():
                self.assertEqual(variables, ComponentSet([model.P[0]]))
            else:
                pred_vars = ComponentSet(
                    [model.rho[i], model.T[i], model.P[i], model.F[i]]
                )
                pred_cons = ComponentSet(
                    [
                        model.ideal_gas[i],
                        model.expansion[i],
                        model.mbal[i],
                        model.ebal[i],
                    ]
                )
                self.assertEqual(pred_vars, variables)
                self.assertEqual(pred_cons, constraints)

    def test_maps_from_triangularization(self):
        N = 5
        model = make_gas_expansion_model(N)
        model.obj = pyo.Objective(expr=0)
        nlp = PyomoNLP(model)
        igraph = IncidenceGraphInterface(nlp)

        # These are the variables and constraints of the square,
        # nonsingular subsystem
        variables = []
        variables.extend(model.P.values())
        variables.extend(
            model.T[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.rho[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.F[i] for i in model.streams if i != model.streams.first()
        )

        constraints = list(model.component_data_objects(pyo.Constraint))

        var_block_map, con_block_map = igraph.map_nodes_to_block_triangular_indices(
            variables, constraints
        )
        var_values = set(var_block_map.values())
        con_values = set(con_block_map.values())
        self.assertEqual(len(var_values), N + 1)
        self.assertEqual(len(con_values), N + 1)

        self.assertEqual(var_block_map[model.P[0]], 0)

        for i in model.streams:
            if i != model.streams.first():
                self.assertEqual(var_block_map[model.rho[i]], i)
                self.assertEqual(var_block_map[model.T[i]], i)
                self.assertEqual(var_block_map[model.P[i]], i)
                self.assertEqual(var_block_map[model.F[i]], i)

                self.assertEqual(con_block_map[model.ideal_gas[i]], i)
                self.assertEqual(con_block_map[model.expansion[i]], i)
                self.assertEqual(con_block_map[model.mbal[i]], i)
                self.assertEqual(con_block_map[model.ebal[i]], i)

    def test_exception(self):
        model = make_gas_expansion_model()
        model.obj = pyo.Objective(expr=0)
        nlp = PyomoNLP(model)
        igraph = IncidenceGraphInterface(nlp)

        with self.assertRaisesRegex(KeyError, "does not exist"):
            variables = [model.P]
            constraints = [model.ideal_gas]
            igraph.maximum_matching(variables, constraints)

        with self.assertRaisesRegex(KeyError, "does not exist"):
            variables = [model.P]
            constraints = [model.ideal_gas]
            igraph.block_triangularize(variables, constraints)


@unittest.skipUnless(networkx_available, "networkx is not available.")
class TestGasExpansionModelInterfaceClassStructural(unittest.TestCase):
    # In these tests we pass a model to the interface and are caching a
    # structural incidence matrix.
    def test_imperfect_matching(self):
        model = make_gas_expansion_model()
        igraph = IncidenceGraphInterface(model)

        n_eqn = len(list(model.component_data_objects(pyo.Constraint)))
        matching = igraph.maximum_matching()
        values = ComponentSet(matching.values())
        self.assertEqual(len(matching), n_eqn)
        self.assertEqual(len(values), n_eqn)

    def test_perfect_matching(self):
        model = make_gas_expansion_model()
        igraph = IncidenceGraphInterface(model)

        # These are the variables and constraints of the square,
        # nonsingular subsystem
        variables = []
        variables.extend(model.P.values())
        variables.extend(
            model.T[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.rho[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.F[i] for i in model.streams if i != model.streams.first()
        )

        constraints = list(model.component_data_objects(pyo.Constraint))

        n_var = len(variables)
        matching = igraph.maximum_matching(variables, constraints)
        values = ComponentSet(matching.values())
        self.assertEqual(len(matching), n_var)
        self.assertEqual(len(values), n_var)

        # The subset of variables and equations we have identified
        # do not have a unique perfect matching. But we at least know
        # this much.
        self.assertIs(matching[model.ideal_gas[0]], model.P[0])

    def test_triangularize(self):
        N = 5
        model = make_gas_expansion_model(N)
        igraph = IncidenceGraphInterface(model)

        # These are the variables and constraints of the square,
        # nonsingular subsystem
        variables = []
        variables.extend(model.P.values())
        variables.extend(
            model.T[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.rho[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.F[i] for i in model.streams if i != model.streams.first()
        )

        constraints = list(model.component_data_objects(pyo.Constraint))

        var_blocks, con_blocks = igraph.block_triangularize(variables, constraints)
        partition = [
            list(zip(vblock, cblock)) for vblock, cblock in zip(var_blocks, con_blocks)
        ]
        self.assertEqual(len(partition), N + 1)

        for i in model.streams:
            variables = ComponentSet([var for var, _ in partition[i]])
            constraints = ComponentSet([con for _, con in partition[i]])
            if i == model.streams.first():
                self.assertEqual(variables, ComponentSet([model.P[0]]))
            else:
                pred_vars = ComponentSet(
                    [model.rho[i], model.T[i], model.P[i], model.F[i]]
                )
                pred_cons = ComponentSet(
                    [
                        model.ideal_gas[i],
                        model.expansion[i],
                        model.mbal[i],
                        model.ebal[i],
                    ]
                )
                self.assertEqual(pred_vars, variables)
                self.assertEqual(pred_cons, constraints)

    def test_maps_from_triangularization(self):
        """
        This tests the maps from variables and constraints to their diagonal
        blocks returned by map_nodes_to_block_triangular_indices
        """
        N = 5
        model = make_gas_expansion_model(N)
        igraph = IncidenceGraphInterface(model)

        # These are the variables and constraints of the square,
        # nonsingular subsystem
        variables = []
        variables.extend(model.P.values())
        variables.extend(
            model.T[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.rho[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.F[i] for i in model.streams if i != model.streams.first()
        )

        constraints = list(model.component_data_objects(pyo.Constraint))

        var_block_map, con_block_map = igraph.map_nodes_to_block_triangular_indices(
            variables, constraints
        )
        var_values = set(var_block_map.values())
        con_values = set(con_block_map.values())
        self.assertEqual(len(var_values), N + 1)
        self.assertEqual(len(con_values), N + 1)

        self.assertEqual(var_block_map[model.P[0]], 0)

        for i in model.streams:
            if i != model.streams.first():
                self.assertEqual(var_block_map[model.rho[i]], i)
                self.assertEqual(var_block_map[model.T[i]], i)
                self.assertEqual(var_block_map[model.P[i]], i)
                self.assertEqual(var_block_map[model.F[i]], i)

                self.assertEqual(con_block_map[model.ideal_gas[i]], i)
                self.assertEqual(con_block_map[model.expansion[i]], i)
                self.assertEqual(con_block_map[model.mbal[i]], i)
                self.assertEqual(con_block_map[model.ebal[i]], i)

    def test_triangularize_submatrix(self):
        # This test exercises triangularization of a somewhat nontrivial
        # submatrix from a cached incidence matrix.
        N = 5
        model = make_gas_expansion_model(N)
        igraph = IncidenceGraphInterface(model)

        # These are the variables and constraints of a square,
        # nonsingular subsystem
        variables = []
        half = N // 2
        variables.extend(model.P[i] for i in model.streams if i >= half)
        variables.extend(model.T[i] for i in model.streams if i > half)
        variables.extend(model.rho[i] for i in model.streams if i > half)
        variables.extend(model.F[i] for i in model.streams if i > half)

        constraints = []
        constraints.extend(model.ideal_gas[i] for i in model.streams if i >= half)
        constraints.extend(model.expansion[i] for i in model.streams if i > half)
        constraints.extend(model.mbal[i] for i in model.streams if i > half)
        constraints.extend(model.ebal[i] for i in model.streams if i > half)

        var_blocks, con_blocks = igraph.block_triangularize(variables, constraints)
        partition = [
            list(zip(vblock, cblock)) for vblock, cblock in zip(var_blocks, con_blocks)
        ]
        self.assertEqual(len(partition), (N - half) + 1)

        for i in model.streams:
            idx = i - half
            variables = ComponentSet([var for var, _ in partition[idx]])
            constraints = ComponentSet([con for _, con in partition[idx]])
            if i == half:
                self.assertEqual(variables, ComponentSet([model.P[half]]))
            elif i > half:
                pred_var = ComponentSet(
                    [model.rho[i], model.T[i], model.P[i], model.F[i]]
                )
                pred_con = ComponentSet(
                    [
                        model.ideal_gas[i],
                        model.expansion[i],
                        model.mbal[i],
                        model.ebal[i],
                    ]
                )
                self.assertEqual(variables, pred_var)
                self.assertEqual(constraints, pred_con)

    def test_maps_from_triangularization_submatrix(self):
        # This test exercises the var/con-block-maps obtained from
        # triangularization of a somewhat nontrivial submatrix from a cached
        # incidence matrix.
        N = 5
        model = make_gas_expansion_model(N)
        igraph = IncidenceGraphInterface(model)

        # These are the variables and constraints of a square,
        # nonsingular subsystem
        variables = []
        half = N // 2
        variables.extend(model.P[i] for i in model.streams if i >= half)
        variables.extend(model.T[i] for i in model.streams if i > half)
        variables.extend(model.rho[i] for i in model.streams if i > half)
        variables.extend(model.F[i] for i in model.streams if i > half)

        constraints = []
        constraints.extend(model.ideal_gas[i] for i in model.streams if i >= half)
        constraints.extend(model.expansion[i] for i in model.streams if i > half)
        constraints.extend(model.mbal[i] for i in model.streams if i > half)
        constraints.extend(model.ebal[i] for i in model.streams if i > half)

        var_block_map, con_block_map = igraph.map_nodes_to_block_triangular_indices(
            variables, constraints
        )
        var_values = set(var_block_map.values())
        con_values = set(con_block_map.values())
        self.assertEqual(len(var_values), (N - half) + 1)
        self.assertEqual(len(con_values), (N - half) + 1)

        self.assertEqual(var_block_map[model.P[half]], 0)

        for i in model.streams:
            if i > half:
                idx = i - half
                self.assertEqual(var_block_map[model.rho[i]], idx)
                self.assertEqual(var_block_map[model.T[i]], idx)
                self.assertEqual(var_block_map[model.P[i]], idx)
                self.assertEqual(var_block_map[model.F[i]], idx)

                self.assertEqual(con_block_map[model.ideal_gas[i]], idx)
                self.assertEqual(con_block_map[model.expansion[i]], idx)
                self.assertEqual(con_block_map[model.mbal[i]], idx)
                self.assertEqual(con_block_map[model.ebal[i]], idx)

    def test_exception(self):
        model = make_gas_expansion_model()
        igraph = IncidenceGraphInterface(model)

        with self.assertRaisesRegex(KeyError, "does not exist"):
            variables = [model.P]
            constraints = [model.ideal_gas]
            igraph.maximum_matching(variables, constraints)

        with self.assertRaisesRegex(KeyError, "does not exist"):
            variables = [model.P]
            constraints = [model.ideal_gas]
            igraph.block_triangularize(variables, constraints)

    @unittest.skipUnless(scipy_available, "scipy is not available.")
    def test_remove(self):
        model = make_gas_expansion_model()
        igraph = IncidenceGraphInterface(model)

        n_eqn = len(list(model.component_data_objects(pyo.Constraint)))
        matching = igraph.maximum_matching()
        values = ComponentSet(matching.values())
        self.assertEqual(len(matching), n_eqn)
        self.assertEqual(len(values), n_eqn)

        variable_set = ComponentSet(igraph.variables)
        self.assertIn(model.F[0], variable_set)
        self.assertIn(model.F[2], variable_set)
        var_dmp, con_dmp = igraph.dulmage_mendelsohn()
        underconstrained_set = ComponentSet(
            var_dmp.unmatched + var_dmp.underconstrained
        )
        self.assertIn(model.F[0], underconstrained_set)
        self.assertIn(model.F[2], underconstrained_set)

        N, M = igraph.incidence_matrix.shape

        # Say we know that these variables and constraints should
        # be matched...
        vars_to_remove = [model.F[0], model.F[2]]
        cons_to_remove = [model.mbal[1], model.mbal[2]]
        igraph.remove_nodes(vars_to_remove, cons_to_remove)
        variable_set = ComponentSet(igraph.variables)
        self.assertNotIn(model.F[0], variable_set)
        self.assertNotIn(model.F[2], variable_set)
        var_dmp, con_dmp = igraph.dulmage_mendelsohn()
        underconstrained_set = ComponentSet(
            var_dmp.unmatched + var_dmp.underconstrained
        )
        self.assertNotIn(model.F[0], underconstrained_set)
        self.assertNotIn(model.F[2], underconstrained_set)

        N_new, M_new = igraph.incidence_matrix.shape
        self.assertEqual(N_new, N - len(cons_to_remove))
        self.assertEqual(M_new, M - len(vars_to_remove))


@unittest.skipUnless(networkx_available, "networkx is not available.")
class TestGasExpansionModelInterfaceClassNoCache(unittest.TestCase):
    # In these tests we do not cache anything and use the interface
    # simply as a convenient wrapper around the analysis functions,
    # which act on matrices.
    def test_imperfect_matching(self):
        model = make_gas_expansion_model()
        igraph = IncidenceGraphInterface()

        constraints = list(model.component_data_objects(pyo.Constraint))
        variables = list(model.component_data_objects(pyo.Var))
        n_eqn = len(constraints)
        matching = igraph.maximum_matching(variables, constraints)
        values = ComponentSet(matching.values())
        self.assertEqual(len(matching), n_eqn)
        self.assertEqual(len(values), n_eqn)

    def test_perfect_matching(self):
        model = make_gas_expansion_model()
        igraph = IncidenceGraphInterface()

        # These are the variables and constraints of the square,
        # nonsingular subsystem
        variables = []
        variables.extend(model.P.values())
        variables.extend(
            model.T[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.rho[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.F[i] for i in model.streams if i != model.streams.first()
        )

        constraints = list(model.component_data_objects(pyo.Constraint))

        n_var = len(variables)
        matching = igraph.maximum_matching(variables, constraints)
        values = ComponentSet(matching.values())
        self.assertEqual(len(matching), n_var)
        self.assertEqual(len(values), n_var)

        # The subset of variables and equations we have identified
        # do not have a unique perfect matching. But we at least know
        # this much.
        self.assertIs(matching[model.ideal_gas[0]], model.P[0])

    def test_triangularize(self):
        N = 5
        model = make_gas_expansion_model(N)
        igraph = IncidenceGraphInterface()

        # These are the variables and constraints of the square,
        # nonsingular subsystem
        variables = []
        variables.extend(model.P.values())
        variables.extend(
            model.T[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.rho[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.F[i] for i in model.streams if i != model.streams.first()
        )

        constraints = list(model.component_data_objects(pyo.Constraint))

        var_blocks, con_blocks = igraph.block_triangularize(variables, constraints)
        partition = [
            list(zip(vblock, cblock)) for vblock, cblock in zip(var_blocks, con_blocks)
        ]
        self.assertEqual(len(partition), N + 1)

        for i in model.streams:
            variables = ComponentSet([var for var, _ in partition[i]])
            constraints = ComponentSet([con for _, con in partition[i]])
            if i == model.streams.first():
                self.assertEqual(variables, ComponentSet([model.P[0]]))
            else:
                pred_vars = ComponentSet(
                    [model.rho[i], model.T[i], model.P[i], model.F[i]]
                )
                pred_cons = ComponentSet(
                    [
                        model.ideal_gas[i],
                        model.expansion[i],
                        model.mbal[i],
                        model.ebal[i],
                    ]
                )
                self.assertEqual(pred_vars, variables)
                self.assertEqual(pred_cons, constraints)

    def test_maps_from_triangularization(self):
        N = 5
        model = make_gas_expansion_model(N)
        igraph = IncidenceGraphInterface()

        # These are the variables and constraints of the square,
        # nonsingular subsystem
        variables = []
        variables.extend(model.P.values())
        variables.extend(
            model.T[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.rho[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.F[i] for i in model.streams if i != model.streams.first()
        )

        constraints = list(model.component_data_objects(pyo.Constraint))

        var_block_map, con_block_map = igraph.map_nodes_to_block_triangular_indices(
            variables, constraints
        )
        var_values = set(var_block_map.values())
        con_values = set(con_block_map.values())
        self.assertEqual(len(var_values), N + 1)
        self.assertEqual(len(con_values), N + 1)

        self.assertEqual(var_block_map[model.P[0]], 0)

        for i in model.streams:
            if i != model.streams.first():
                self.assertEqual(var_block_map[model.rho[i]], i)
                self.assertEqual(var_block_map[model.T[i]], i)
                self.assertEqual(var_block_map[model.P[i]], i)
                self.assertEqual(var_block_map[model.F[i]], i)

                self.assertEqual(con_block_map[model.ideal_gas[i]], i)
                self.assertEqual(con_block_map[model.expansion[i]], i)
                self.assertEqual(con_block_map[model.mbal[i]], i)
                self.assertEqual(con_block_map[model.ebal[i]], i)

    def test_diagonal_blocks(self):
        N = 5
        model = make_gas_expansion_model(N)
        igraph = IncidenceGraphInterface()

        # These are the variables and constraints of the square,
        # nonsingular subsystem
        variables = []
        variables.extend(model.P.values())
        variables.extend(
            model.T[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.rho[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.F[i] for i in model.streams if i != model.streams.first()
        )

        constraints = list(model.component_data_objects(pyo.Constraint))

        var_blocks, con_blocks = igraph.get_diagonal_blocks(variables, constraints)
        # self.assertIs(igraph.row_block_map, None)
        # self.assertIs(igraph.col_block_map, None)
        self.assertEqual(len(var_blocks), N + 1)
        self.assertEqual(len(con_blocks), N + 1)

        for i, (vars, cons) in enumerate(zip(var_blocks, con_blocks)):
            var_set = ComponentSet(vars)
            con_set = ComponentSet(cons)

            if i == 0:
                pred_var_set = ComponentSet([model.P[0]])
                self.assertEqual(pred_var_set, var_set)
                pred_con_set = ComponentSet([model.ideal_gas[0]])
                self.assertEqual(pred_con_set, con_set)

            else:
                pred_var_set = ComponentSet(
                    [model.rho[i], model.T[i], model.P[i], model.F[i]]
                )
                pred_con_set = ComponentSet(
                    [
                        model.ideal_gas[i],
                        model.expansion[i],
                        model.mbal[i],
                        model.ebal[i],
                    ]
                )
                self.assertEqual(pred_var_set, var_set)
                self.assertEqual(pred_con_set, con_set)

    def test_diagonal_blocks_with_cached_maps(self):
        # NOTE: This functionality has been deprecated.
        N = 5
        model = make_gas_expansion_model(N)
        igraph = IncidenceGraphInterface()

        # These are the variables and constraints of the square,
        # nonsingular subsystem
        variables = []
        variables.extend(model.P.values())
        variables.extend(
            model.T[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.rho[i] for i in model.streams if i != model.streams.first()
        )
        variables.extend(
            model.F[i] for i in model.streams if i != model.streams.first()
        )

        constraints = list(model.component_data_objects(pyo.Constraint))

        igraph.block_triangularize(variables, constraints)
        var_blocks, con_blocks = igraph.get_diagonal_blocks(variables, constraints)
        # NOTE: row/col_block_map have been deprecated.
        # However, they still return None for now.
        self.assertIs(igraph.row_block_map, None)
        self.assertIs(igraph.col_block_map, None)


@unittest.skipUnless(networkx_available, "networkx is not available.")
class TestDulmageMendelsohnInterface(unittest.TestCase):
    def test_degenerate_solid_phase_model(self):
        m = make_degenerate_solid_phase_model()
        variables = list(m.component_data_objects(pyo.Var))
        constraints = list(m.component_data_objects(pyo.Constraint))

        igraph = IncidenceGraphInterface()
        var_dmp, con_dmp = igraph.dulmage_mendelsohn(variables, constraints)

        underconstrained_vars = ComponentSet(m.flow_comp.values())
        underconstrained_vars.add(m.flow)
        underconstrained_cons = ComponentSet(m.flow_eqn.values())

        self.assertEqual(len(var_dmp[0] + var_dmp[1]), len(underconstrained_vars))
        for var in var_dmp[0] + var_dmp[1]:
            self.assertIn(var, underconstrained_vars)

        self.assertEqual(len(con_dmp[2]), len(underconstrained_cons))
        for con in con_dmp[2]:
            self.assertIn(con, underconstrained_cons)

        overconstrained_cons = ComponentSet(m.holdup_eqn.values())
        overconstrained_cons.add(m.density_eqn)
        overconstrained_cons.add(m.sum_eqn)
        overconstrained_vars = ComponentSet(m.x.values())
        overconstrained_vars.add(m.rho)

        self.assertEqual(len(var_dmp[2]), len(overconstrained_vars))
        for var in var_dmp[2]:
            self.assertIn(var, overconstrained_vars)

        self.assertEqual(len(con_dmp[0] + con_dmp[1]), len(overconstrained_cons))
        for con in con_dmp[0] + con_dmp[1]:
            self.assertIn(con, overconstrained_cons)

    def test_named_tuple(self):
        m = make_degenerate_solid_phase_model()
        variables = list(m.component_data_objects(pyo.Var))
        constraints = list(m.component_data_objects(pyo.Constraint))

        igraph = IncidenceGraphInterface()
        var_dmp, con_dmp = igraph.dulmage_mendelsohn(variables, constraints)

        underconstrained_vars = ComponentSet(m.flow_comp.values())
        underconstrained_vars.add(m.flow)
        underconstrained_cons = ComponentSet(m.flow_eqn.values())

        dmp_vars_under = var_dmp.unmatched + var_dmp.underconstrained
        dmp_vars_over = var_dmp.overconstrained
        dmp_cons_under = con_dmp.underconstrained
        dmp_cons_over = con_dmp.unmatched + con_dmp.overconstrained

        self.assertEqual(len(dmp_vars_under), len(underconstrained_vars))
        for var in dmp_vars_under:
            self.assertIn(var, underconstrained_vars)

        self.assertEqual(len(dmp_cons_under), len(underconstrained_cons))
        for con in dmp_cons_under:
            self.assertIn(con, underconstrained_cons)

        overconstrained_cons = ComponentSet(m.holdup_eqn.values())
        overconstrained_cons.add(m.density_eqn)
        overconstrained_cons.add(m.sum_eqn)
        overconstrained_vars = ComponentSet(m.x.values())
        overconstrained_vars.add(m.rho)

        self.assertEqual(len(dmp_vars_over), len(overconstrained_vars))
        for var in dmp_vars_over:
            self.assertIn(var, overconstrained_vars)

        self.assertEqual(len(dmp_cons_over), len(overconstrained_cons))
        for con in dmp_cons_over:
            self.assertIn(con, overconstrained_cons)

    @unittest.skipUnless(scipy_available, "scipy is not available.")
    def test_incidence_graph(self):
        m = make_degenerate_solid_phase_model()
        variables = list(m.component_data_objects(pyo.Var))
        constraints = list(m.component_data_objects(pyo.Constraint))
        graph = get_incidence_graph(variables, constraints)
        matrix = get_structural_incidence_matrix(variables, constraints)
        from_matrix = from_biadjacency_matrix(matrix)

        self.assertEqual(graph.nodes, from_matrix.nodes)
        self.assertEqual(graph.edges, from_matrix.edges)

    def test_dm_graph_interface(self):
        m = make_degenerate_solid_phase_model()
        variables = list(m.component_data_objects(pyo.Var))
        constraints = list(m.component_data_objects(pyo.Constraint))
        graph = get_incidence_graph(variables, constraints)

        M, N = len(constraints), len(variables)

        top_nodes = list(range(M))
        con_dmp, var_dmp = dulmage_mendelsohn(graph, top_nodes=top_nodes)
        con_dmp = tuple([constraints[i] for i in subset] for subset in con_dmp)
        var_dmp = tuple([variables[i - M] for i in subset] for subset in var_dmp)

        underconstrained_vars = ComponentSet(m.flow_comp.values())
        underconstrained_vars.add(m.flow)
        underconstrained_cons = ComponentSet(m.flow_eqn.values())

        self.assertEqual(len(var_dmp[0] + var_dmp[1]), len(underconstrained_vars))
        for var in var_dmp[0] + var_dmp[1]:
            self.assertIn(var, underconstrained_vars)

        self.assertEqual(len(con_dmp[2]), len(underconstrained_cons))
        for con in con_dmp[2]:
            self.assertIn(con, underconstrained_cons)

        overconstrained_cons = ComponentSet(m.holdup_eqn.values())
        overconstrained_cons.add(m.density_eqn)
        overconstrained_cons.add(m.sum_eqn)
        overconstrained_vars = ComponentSet(m.x.values())
        overconstrained_vars.add(m.rho)

        self.assertEqual(len(var_dmp[2]), len(overconstrained_vars))
        for var in var_dmp[2]:
            self.assertIn(var, overconstrained_vars)

        self.assertEqual(len(con_dmp[0] + con_dmp[1]), len(overconstrained_cons))
        for con in con_dmp[0] + con_dmp[1]:
            self.assertIn(con, overconstrained_cons)

    @unittest.skipUnless(scipy_available, "scipy is not available.")
    def test_remove(self):
        m = make_degenerate_solid_phase_model()
        variables = list(m.component_data_objects(pyo.Var))
        constraints = list(m.component_data_objects(pyo.Constraint))

        igraph = IncidenceGraphInterface(m)
        var_dmp, con_dmp = igraph.dulmage_mendelsohn()
        var_con_set = ComponentSet(igraph.variables + igraph.constraints)
        underconstrained_set = ComponentSet(
            var_dmp.unmatched + var_dmp.underconstrained
        )
        self.assertIn(m.flow_comp[1], var_con_set)
        self.assertIn(m.flow_eqn[1], var_con_set)
        self.assertIn(m.flow_comp[1], underconstrained_set)

        N, M = igraph.incidence_matrix.shape

        # flow_comp[1] is underconstrained, but we think it should be
        # specified by flow_eqn[1], so we remove these from the incidence
        # matrix.
        vars_to_remove = [m.flow_comp[1]]
        cons_to_remove = [m.flow_eqn[1]]
        igraph.remove_nodes(vars_to_remove, cons_to_remove)
        var_dmp, con_dmp = igraph.dulmage_mendelsohn()
        var_con_set = ComponentSet(igraph.variables + igraph.constraints)
        underconstrained_set = ComponentSet(
            var_dmp.unmatched + var_dmp.underconstrained
        )
        self.assertNotIn(m.flow_comp[1], var_con_set)
        self.assertNotIn(m.flow_eqn[1], var_con_set)
        self.assertNotIn(m.flow_comp[1], underconstrained_set)

        N_new, M_new = igraph.incidence_matrix.shape
        self.assertEqual(N_new, N - len(cons_to_remove))
        self.assertEqual(M_new, M - len(vars_to_remove))

    def test_recover_matching_from_dulmage_mendelsohn(self):
        m = make_degenerate_solid_phase_model()
        igraph = IncidenceGraphInterface(m)
        vdmp, cdmp = igraph.dulmage_mendelsohn()
        vmatch = vdmp.underconstrained + vdmp.square + vdmp.overconstrained
        cmatch = cdmp.underconstrained + cdmp.square + cdmp.overconstrained
        # Assert no duplicates in matched variables and constraints
        self.assertEqual(len(ComponentSet(vmatch)), len(vmatch))
        self.assertEqual(len(ComponentSet(cmatch)), len(cmatch))
        matching = list(zip(vmatch, cmatch))
        # Assert each matched pair contains a variable that participates
        # in the constraint.
        for var, con in matching:
            var_in_con = ComponentSet(igraph.get_adjacent_to(con))
            self.assertIn(var, var_in_con)


@unittest.skipUnless(networkx_available, "networkx is not available.")
class TestConnectedComponents(unittest.TestCase):
    def test_dynamic_model_backward(self):
        """
        This is the same test as performed in the test_connected.py
        file, now implemented with the Pyomo interface.
        """
        m = make_dynamic_model(nfe=5, scheme="BACKWARD")
        m.height[0].fix()
        igraph = IncidenceGraphInterface(m)
        var_blocks, con_blocks = igraph.get_connected_components()
        vc_blocks = [
            (tuple(vars), tuple(cons)) for vars, cons in zip(var_blocks, con_blocks)
        ]
        key_fcn = lambda vc_comps: tuple(
            tuple(comp.name for comp in comps) for comps in vc_comps
        )
        vc_blocks = list(sorted(vc_blocks, key=key_fcn))

        t0_vars = ComponentSet((m.flow_out[0], m.dhdt[0], m.flow_in[0]))
        t0_cons = ComponentSet((m.flow_out_eqn[0], m.diff_eqn[0]))

        # The variables in these blocks need to be sorted by their coordinates
        # in the underlying incidence matrix
        var_key = lambda var: igraph.get_matrix_coord(var)
        con_key = lambda con: igraph.get_matrix_coord(con)
        var_blocks = [
            tuple(sorted(t0_vars, key=var_key)),
            tuple(
                sorted(
                    (var for var in igraph.variables if var not in t0_vars), key=var_key
                )
            ),
        ]
        con_blocks = [
            tuple(sorted(t0_cons, key=con_key)),
            tuple(
                sorted(
                    (con for con in igraph.constraints if con not in t0_cons),
                    key=con_key,
                )
            ),
        ]
        target_blocks = [
            (tuple(vars), tuple(cons)) for vars, cons in zip(var_blocks, con_blocks)
        ]
        target_blocks = list(sorted(target_blocks, key=key_fcn))

        # I am somewhat surprised this works. This appears to because
        # var1 == var2 is a constant equality expression when var1 is var2.
        # So if this test fails, we'll get a somewhat confusing PyomoException
        # about not being able to convert non-constant expressions to bool
        # rather than a message saying that our variables are not the same.
        # self.assertEqual(target_blocks, vc_blocks)
        for block, target_block in zip(vc_blocks, target_blocks):
            vars, cons = block
            pred_vars, pred_cons = target_block
            self.assertEqual(len(vars), len(pred_vars))
            self.assertEqual(len(cons), len(pred_cons))
            for v1, v2 in zip(vars, pred_vars):
                self.assertIs(v1, v2)
            for c1, c2 in zip(cons, pred_cons):
                self.assertIs(c1, c2)


@unittest.skipUnless(networkx_available, "networkx is not available.")
@unittest.skipUnless(scipy_available, "scipy is not available.")
class TestExtraVars(unittest.TestCase):
    def test_unused_var(self):
        m = pyo.ConcreteModel()
        m.v1 = pyo.Var()
        m.v2 = pyo.Var()
        m.c1 = pyo.Constraint(expr=m.v1 == 1.0)
        igraph = IncidenceGraphInterface(m)
        self.assertEqual(igraph.incidence_matrix.shape, (1, 1))

    def test_reference(self):
        m = pyo.ConcreteModel()
        m.v1 = pyo.Var()
        m.ref = pyo.Reference(m.v1)
        m.c1 = pyo.Constraint(expr=m.v1 == 1.0)
        igraph = IncidenceGraphInterface(m)
        self.assertEqual(igraph.incidence_matrix.shape, (1, 1))


@unittest.skipUnless(networkx_available, "networkx is not available.")
class TestExceptions(unittest.TestCase):
    @unittest.skipUnless(scipy_available, "scipy is not available.")
    @unittest.skipUnless(asl_available, "pynumero_ASL is not available")
    def test_nlp_fixed_error(self):
        m = pyo.ConcreteModel()
        m.v1 = pyo.Var()
        m.v2 = pyo.Var()
        m.c1 = pyo.Constraint(expr=m.v1 + m.v2 == 1.0)
        m.v2.fix(2.0)
        m._obj = pyo.Objective(expr=0.0)
        nlp = PyomoNLP(m)
        msg = "generation options.*are not supported"
        with self.assertRaisesRegex(ValueError, msg):
            igraph = IncidenceGraphInterface(nlp, include_fixed=True)

    @unittest.skipUnless(scipy_available, "scipy is not available.")
    @unittest.skipUnless(asl_available, "pynumero_ASL is not available")
    def test_nlp_active_error(self):
        m = pyo.ConcreteModel()
        m.v1 = pyo.Var()
        m.c1 = pyo.Constraint(expr=m.v1 == 1.0)
        m.c2 = pyo.Constraint(expr=m.v1 == 2.0)
        m._obj = pyo.Objective(expr=0.0)
        nlp = PyomoNLP(m)
        with self.assertRaisesRegex(ValueError, "inactive constraints"):
            igraph = IncidenceGraphInterface(nlp, active=False)

    def test_remove_no_matrix(self):
        m = pyo.ConcreteModel()
        m.v1 = pyo.Var()
        igraph = IncidenceGraphInterface()
        with self.assertRaisesRegex(RuntimeError, "no incidence matrix"):
            igraph.remove_nodes([m.v1])

    def test_remove_bad_node(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3])
        m.eq = pyo.Constraint(pyo.PositiveIntegers)
        m.eq[1] = m.x[1] * m.x[2] == m.x[3]
        m.eq[2] = m.x[1] + 2 * m.x[2] == 3 * m.x[3]
        igraph = IncidenceGraphInterface(m)
        with self.assertRaisesRegex(KeyError, "does not exist"):
            # Suppose we think something like this should work. We should get
            # an error, and not silently do nothing.
            igraph.remove_nodes([m.x], [m.eq[1]])

        with self.assertRaisesRegex(KeyError, "does not exist"):
            igraph.remove_nodes(None, [m.eq])

        with self.assertRaisesRegex(KeyError, "does not exist"):
            igraph.remove_nodes([[m.x[1], m.x[2]], [m.eq[1]]])

    def test_remove_varcon_samelist_deprecated(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3])
        m.eq = pyo.Constraint(pyo.PositiveIntegers)
        m.eq[1] = m.x[1] * m.x[2] == m.x[3]
        m.eq[2] = m.x[1] + 2 * m.x[2] == 3 * m.x[3]

        igraph = IncidenceGraphInterface(m)
        # This raises a deprecation warning. When the deprecated functionality
        # is removed, this will fail, and this test should be updated accordingly.
        igraph.remove_nodes([m.eq[1], m.x[1]])
        self.assertEqual(len(igraph.variables), 2)
        self.assertEqual(len(igraph.constraints), 1)

        igraph.remove_nodes([], [m.eq[2], m.x[2]])
        self.assertEqual(len(igraph.variables), 1)
        self.assertEqual(len(igraph.constraints), 0)


@unittest.skipUnless(networkx_available, "networkx is not available.")
@unittest.skipUnless(scipy_available, "scipy is not available.")
class TestIncludeInequality(unittest.TestCase):
    def make_model_with_inequalities(self):
        m = make_degenerate_solid_phase_model()

        @m.Constraint()
        def flow_bound(m):
            return m.flow >= 0

        @m.Constraint(m.components)
        def flow_comp_bound(m, j):
            return m.flow_comp[j] >= 0

        return m

    def test_dont_include_inequality_model(self):
        m = self.make_model_with_inequalities()
        igraph = IncidenceGraphInterface(m, include_inequality=False)
        self.assertEqual(igraph.incidence_matrix.shape, (8, 8))

    def test_include_inequality_model(self):
        m = self.make_model_with_inequalities()
        igraph = IncidenceGraphInterface(m, include_inequality=True)
        self.assertEqual(igraph.incidence_matrix.shape, (12, 8))

    @unittest.skipUnless(asl_available, "pynumero_ASL is not available")
    def test_dont_include_inequality_nlp(self):
        m = self.make_model_with_inequalities()
        m._obj = pyo.Objective(expr=0)
        nlp = PyomoNLP(m)
        igraph = IncidenceGraphInterface(nlp, include_inequality=False)
        self.assertEqual(igraph.incidence_matrix.shape, (8, 8))

    @unittest.skipUnless(asl_available, "pynumero_ASL is not available")
    def test_include_inequality_nlp(self):
        m = self.make_model_with_inequalities()
        m._obj = pyo.Objective(expr=0)
        nlp = PyomoNLP(m)
        igraph = IncidenceGraphInterface(nlp, include_inequality=True)
        self.assertEqual(igraph.incidence_matrix.shape, (12, 8))


@unittest.skipUnless(networkx_available, "networkx is not available.")
class TestGetIncidenceGraph(unittest.TestCase):
    def make_test_model(self):
        m = pyo.ConcreteModel()
        m.I = pyo.Set(initialize=[1, 2, 3, 4])
        m.v = pyo.Var(m.I, bounds=(0, None))
        m.eq1 = pyo.Constraint(expr=m.v[1] ** 2 + m.v[2] ** 2 == 1.0)
        m.eq2 = pyo.Constraint(expr=m.v[1] + 2.0 == m.v[3])
        m.ineq1 = pyo.Constraint(expr=m.v[2] - m.v[3] ** 0.5 + m.v[4] ** 2 <= 1.0)
        m.ineq2 = pyo.Constraint(expr=m.v[2] * m.v[4] >= 1.0)
        m.ineq3 = pyo.Constraint(expr=m.v[1] >= m.v[4] ** 4)
        m.obj = pyo.Objective(expr=-m.v[1] - m.v[2] + m.v[3] ** 2 + m.v[4] ** 2)
        return m

    def test_bipartite_incidence_graph(self):
        m = self.make_test_model()
        constraints = [m.eq1, m.eq2, m.ineq1, m.ineq2, m.ineq3]
        variables = list(m.v.values())
        graph = get_bipartite_incidence_graph(variables, constraints)

        # Nodes:
        #   0: m.eq1
        #   1: m.eq2
        #   2: m.ineq1
        #   3: m.ineq2
        #   4: m.ineq3
        #   5: m.v[1]
        #   6: m.v[2]
        #   7: m.v[3]
        #   8: m.v[4]

        # Assert some basic structure
        self.assertEqual(len(graph.nodes), 9)
        self.assertEqual(len(graph.edges), 11)
        self.assertTrue(nx.algorithms.bipartite.is_bipartite(graph))

        # Assert that the "adjacency list" is what we expect
        self.assertEqual(set(graph[0]), {5, 6})
        self.assertEqual(set(graph[1]), {5, 7})
        self.assertEqual(set(graph[2]), {6, 7, 8})
        self.assertEqual(set(graph[3]), {6, 8})
        self.assertEqual(set(graph[4]), {5, 8})
        self.assertEqual(set(graph[5]), {0, 1, 4})
        self.assertEqual(set(graph[6]), {0, 2, 3})
        self.assertEqual(set(graph[7]), {1, 2})
        self.assertEqual(set(graph[8]), {2, 3, 4})

    def test_unused_var(self):
        m = self.make_test_model()
        constraints = [m.eq1, m.eq2]
        variables = list(m.v.values())
        graph = get_bipartite_incidence_graph(variables, constraints)

        # Nodes:
        #   0: m.eq1
        #   1: m.eq2
        #   2: m.v[1]
        #   3: m.v[2]
        #   4: m.v[3]
        #   5: m.v[4]

        self.assertEqual(len(graph.nodes), 6)
        self.assertEqual(len(graph.edges), 4)
        self.assertTrue(nx.algorithms.bipartite.is_bipartite(graph))

        # Assert that the "adjacency list" is what we expect
        self.assertEqual(set(graph[0]), {2, 3})
        self.assertEqual(set(graph[1]), {2, 4})
        self.assertEqual(set(graph[2]), {0, 1})
        self.assertEqual(set(graph[3]), {0})
        self.assertEqual(set(graph[4]), {1})
        self.assertEqual(set(graph[5]), set())

    def test_fixed_vars(self):
        m = self.make_test_model()
        constraints = [m.eq1, m.eq2, m.ineq1, m.ineq2, m.ineq3]
        variables = list(m.v.values())
        m.v[1].fix()
        m.v[4].fix()

        # Slightly odd situation where we provide fixed variables, but
        # then tell the graph to not include them. Nodes will be created
        # for these vars, but they will not have any edges.
        graph = get_bipartite_incidence_graph(
            variables, constraints, include_fixed=False
        )

        # Nodes:
        #   0: m.eq1
        #   1: m.eq2
        #   2: m.ineq1
        #   3: m.ineq2
        #   4: m.ineq3
        #   5: m.v[1]
        #   6: m.v[2]
        #   7: m.v[3]
        #   8: m.v[4]

        self.assertEqual(len(graph.nodes), 9)
        self.assertEqual(len(graph.edges), 5)
        self.assertTrue(nx.algorithms.bipartite.is_bipartite(graph))

        # Assert that the "adjacency list" is what we expect
        self.assertEqual(set(graph[0]), {6})
        self.assertEqual(set(graph[1]), {7})
        self.assertEqual(set(graph[2]), {6, 7})
        self.assertEqual(set(graph[3]), {6})
        self.assertEqual(set(graph[4]), set())
        self.assertEqual(set(graph[5]), set())
        self.assertEqual(set(graph[6]), {0, 2, 3})
        self.assertEqual(set(graph[7]), {1, 2})
        self.assertEqual(set(graph[8]), set())

    def test_extract_subgraph(self):
        m = self.make_test_model()
        constraints = [m.eq1, m.eq2, m.ineq1, m.ineq2, m.ineq3]
        variables = list(m.v.values())
        graph = get_bipartite_incidence_graph(variables, constraints)

        sg_cons = [0, 2]
        sg_vars = [i + len(constraints) for i in [2, 0, 3]]

        subgraph = extract_bipartite_subgraph(graph, sg_cons, sg_vars)

        # Subgraph nodes:
        #   0: m.eq1
        #   1: m.ineq1
        #   2: m.v[3]
        #   3: m.v[1]
        #   4: m.v[4]

        self.assertEqual(len(subgraph.nodes), 5)
        self.assertEqual(len(subgraph.edges), 3)
        self.assertTrue(nx.algorithms.bipartite.is_bipartite(subgraph))

        self.assertEqual(set(subgraph[0]), {3})
        self.assertEqual(set(subgraph[1]), {2, 4})
        self.assertEqual(set(subgraph[2]), {1})
        self.assertEqual(set(subgraph[3]), {0})
        self.assertEqual(set(subgraph[4]), {1})

    def test_extract_exceptions(self):
        m = self.make_test_model()
        constraints = [m.eq1, m.eq2, m.ineq1, m.ineq2, m.ineq3]
        variables = list(m.v.values())
        graph = get_bipartite_incidence_graph(variables, constraints)

        sg_cons = [0, 2, 5]
        sg_vars = [i + len(constraints) for i in [2, 3]]
        msg = "Invalid bipartite sets."
        with self.assertRaisesRegex(RuntimeError, msg):
            subgraph = extract_bipartite_subgraph(graph, sg_cons, sg_vars)

        sg_cons = [0, 2, 0]
        sg_vars = [i + len(constraints) for i in [2, 0, 3]]
        msg = "provided more than once"
        with self.assertRaisesRegex(RuntimeError, msg):
            subgraph = extract_bipartite_subgraph(graph, sg_cons, sg_vars)


@unittest.skipUnless(networkx_available, "networkx is not available.")
class TestGetAdjacent(unittest.TestCase):
    def test_get_adjacent_to_var(self):
        m = make_degenerate_solid_phase_model()
        igraph = IncidenceGraphInterface(m)
        adj_cons = igraph.get_adjacent_to(m.rho)
        self.assertEqual(
            ComponentSet(adj_cons),
            ComponentSet(
                [m.holdup_eqn[1], m.holdup_eqn[2], m.holdup_eqn[3], m.density_eqn]
            ),
        )

    def test_get_adjacent_to_con(self):
        m = make_degenerate_solid_phase_model()
        igraph = IncidenceGraphInterface(m)
        adj_vars = igraph.get_adjacent_to(m.density_eqn)
        self.assertEqual(
            ComponentSet(adj_vars), ComponentSet([m.x[1], m.x[2], m.x[3], m.rho])
        )

    def test_get_adjacent_exceptions(self):
        m = make_degenerate_solid_phase_model()
        igraph = IncidenceGraphInterface()
        msg = "Cannot get components adjacent to"
        with self.assertRaisesRegex(RuntimeError, msg):
            adj_vars = igraph.get_adjacent_to(m.density_eqn)

        m.x[1].fix()
        igraph = IncidenceGraphInterface(m, include_fixed=False)
        msg = "Cannot find component"
        with self.assertRaisesRegex(RuntimeError, msg):
            adj_cons = igraph.get_adjacent_to(m.x[1])


@unittest.skipUnless(networkx_available, "networkx is not available.")
class TestInterface(unittest.TestCase):
    def test_assumed_constraint_behavior(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3])
        m.con = pyo.Constraint(expr=m.x[1] == m.x[2] - pyo.exp(m.x[3]))
        var_set = ComponentSet(identify_variables(m.con.body))
        self.assertEqual(var_set, ComponentSet(m.x[:]))

    def test_subgraph_with_fewer_var_or_con(self):
        m = pyo.ConcreteModel()
        m.I = pyo.Set(initialize=[1, 2])
        m.v = pyo.Var(m.I)
        m.eq1 = pyo.Constraint(expr=m.v[1] + m.v[2] == 1)
        m.ineq1 = pyo.Constraint(expr=m.v[1] - m.v[2] <= 2)

        # Defensively set include_inequality=True, which is the current
        # default, in case this default changes.
        igraph = IncidenceGraphInterface(m, include_inequality=True)

        variables = list(m.v.values())
        constraints = [m.ineq1]
        matching = igraph.maximum_matching(variables, constraints)
        self.assertEqual(len(matching), 1)

        variables = [m.v[2]]
        constraints = [m.eq1, m.ineq1]
        matching = igraph.maximum_matching(variables, constraints)
        self.assertEqual(len(matching), 1)

    @unittest.skipUnless(plotly_available, "Plotly is not available")
    def test_plot(self):
        """
        Unfortunately, this test only ensures the code runs without errors.
        It does not test for correctness.
        """
        m = pyo.ConcreteModel()
        m.x = pyo.Var(bounds=(-1, 1))
        m.y = pyo.Var()
        m.z = pyo.Var()
        # NOTE: Objective will not be displayed
        m.obj = pyo.Objective(expr=m.y**2 + m.z**2)
        m.c1 = pyo.Constraint(expr=m.y == 2 * m.x + 1)
        m.c2 = pyo.Constraint(expr=m.z >= m.x)
        m.y.fix()
        igraph = IncidenceGraphInterface(m, include_inequality=True, include_fixed=True)
        igraph.plot(title="test plot", show=False)

    def test_zero_coeff(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3])
        m.eq1 = pyo.Constraint(expr=m.x[1] + 0 * m.x[2] == 2)
        m.eq2 = pyo.Constraint(expr=m.x[1] ** 2 == 1)
        m.eq3 = pyo.Constraint(expr=m.x[2] * m.x[3] - m.x[1] == 1)

        igraph = IncidenceGraphInterface(m)
        var_dmp, con_dmp = igraph.dulmage_mendelsohn()

        # Because 0*m.x[2] does not appear in the incidence graph, we correctly
        # identify that the system is structurally singular
        self.assertGreater(len(var_dmp.unmatched), 0)

    def test_var_minus_itself(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3])
        m.eq1 = pyo.Constraint(expr=m.x[1] + m.x[2] - m.x[2] == 2)
        m.eq2 = pyo.Constraint(expr=m.x[1] ** 2 == 1)
        m.eq3 = pyo.Constraint(expr=m.x[2] * m.x[3] - m.x[1] == 1)

        igraph = IncidenceGraphInterface(m)
        var_dmp, con_dmp = igraph.dulmage_mendelsohn()

        # m.x[2] - m.x[2] is correctly ignored by generate_standard_repn,
        # so we correctly identify that the system is structurally singular
        self.assertGreater(len(var_dmp.unmatched), 0)

    def test_linear_only(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3])
        m.eq1 = pyo.Constraint(expr=m.x[1] ** 2 + m.x[2] ** 2 + m.x[3] ** 2 == 1)
        m.eq2 = pyo.Constraint(expr=m.x[2] + pyo.sqrt(m.x[1]) + pyo.exp(m.x[3]) == 1)
        m.eq3 = pyo.Constraint(expr=m.x[3] + m.x[1] ** 3 + m.x[2] == 1)

        igraph = IncidenceGraphInterface(m, linear_only=True)
        self.assertEqual(igraph.n_edges, 3)
        self.assertEqual(ComponentSet(igraph.variables), ComponentSet([m.x[2], m.x[3]]))

        matching = igraph.maximum_matching()
        self.assertEqual(len(matching), 2)
        self.assertIs(matching[m.eq2], m.x[2])
        self.assertIs(matching[m.eq3], m.x[3])

    def test_add_edge(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3, 4])
        m.eq1 = pyo.Constraint(expr=m.x[1] ** 2 + m.x[2] ** 2 + m.x[3] ** 2 == 1)
        m.eq2 = pyo.Constraint(expr=m.x[2] + pyo.sqrt(m.x[1]) + pyo.exp(m.x[3]) == 1)
        m.eq3 = pyo.Constraint(expr=m.x[3] + m.x[2] + m.x[4] == 1)
        m.eq4 = pyo.Constraint(expr=m.x[1] + m.x[2] ** 2 == 5)

        igraph = IncidenceGraphInterface(m, linear_only=False)
        n_edges_original = igraph.n_edges

        # Test edge is added between previously unconnected nodes
        igraph.add_edge(m.x[1], m.eq3)
        n_edges_new = igraph.n_edges
        assert ComponentSet(igraph.get_adjacent_to(m.eq3)) == ComponentSet(m.x[:])
        self.assertEqual(n_edges_original + 1, n_edges_new)

        # Test no edge is added if there exists a previous edge between nodes
        igraph.add_edge(m.x[2], m.eq3)
        n_edges2 = igraph.n_edges
        self.assertEqual(n_edges_new, n_edges2)

    def test_add_edge_linear_igraph(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3, 4])
        m.eq1 = pyo.Constraint(expr=m.x[1] + m.x[3] == 1)
        m.eq2 = pyo.Constraint(expr=m.x[2] + pyo.sqrt(m.x[1]) + pyo.exp(m.x[3]) == 1)
        m.eq3 = pyo.Constraint(expr=m.x[4] ** 2 + m.x[1] ** 3 + m.x[2] ** 2 == 1)

        # Make sure error is raised when a variable is not in the igraph
        igraph = IncidenceGraphInterface(m, linear_only=True)

        msg = "is not a variable in the incidence graph"
        with self.assertRaisesRegex(RuntimeError, msg):
            igraph.add_edge(m.x[4], m.eq2)

    def test_var_elim(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3, 4])
        m.eq1 = pyo.Constraint(expr=m.x[1] ** 2 + m.x[2] ** 2 + m.x[3] ** 2 == 1)
        m.eq2 = pyo.Constraint(expr=pyo.sqrt(m.x[1]) + pyo.exp(m.x[3]) == 1)
        m.eq3 = pyo.Constraint(expr=m.x[3] + m.x[2] + m.x[4] == 1)
        m.eq4 = pyo.Constraint(expr=m.x[1] == 5 * m.x[2])

        igraph = IncidenceGraphInterface(m)
        # Eliminate x[1] using eq4
        for adj_con in igraph.get_adjacent_to(m.x[1]):
            for adj_var in igraph.get_adjacent_to(m.eq4):
                igraph.add_edge(adj_var, adj_con)
        igraph.remove_nodes([m.x[1]], [m.eq4])

        assert ComponentSet(igraph.variables) == ComponentSet([m.x[2], m.x[3], m.x[4]])
        assert ComponentSet(igraph.constraints) == ComponentSet([m.eq1, m.eq2, m.eq3])
        self.assertEqual(7, igraph.n_edges)

        assert m.x[2] in ComponentSet(igraph.get_adjacent_to(m.eq1))
        assert m.x[2] in ComponentSet(igraph.get_adjacent_to(m.eq2))

    def test_subgraph(self):
        m = pyo.ConcreteModel()
        m.I = pyo.Set(initialize=[1, 2, 3, 4])
        m.v = pyo.Var(m.I, bounds=(0, None))
        m.eq1 = pyo.Constraint(expr=m.v[1] ** 2 + m.v[2] ** 2 == 1.0)
        m.eq2 = pyo.Constraint(expr=m.v[1] + 2.0 == m.v[3])
        m.ineq1 = pyo.Constraint(expr=m.v[2] - m.v[3] ** 0.5 + m.v[4] ** 2 <= 1.0)
        m.ineq2 = pyo.Constraint(expr=m.v[2] * m.v[4] >= 1.0)
        m.ineq3 = pyo.Constraint(expr=m.v[1] >= m.v[4] ** 4)
        m.obj = pyo.Objective(expr=-m.v[1] - m.v[2] + m.v[3] ** 2 + m.v[4] ** 2)
        igraph = IncidenceGraphInterface(m)
        eq_igraph = igraph.subgraph(igraph.variables, [m.eq1, m.eq2])
        for i in range(len(igraph.variables)):
            self.assertIs(igraph.variables[i], eq_igraph.variables[i])
        self.assertEqual(
            ComponentSet(eq_igraph.constraints), ComponentSet([m.eq1, m.eq2])
        )

        subgraph = eq_igraph.subgraph([m.v[1], m.v[3]], [m.eq1, m.eq2])
        self.assertEqual(
            ComponentSet(subgraph.get_adjacent_to(m.eq2)),
            ComponentSet([m.v[1], m.v[3]]),
        )
        self.assertEqual(
            ComponentSet(subgraph.get_adjacent_to(m.eq1)), ComponentSet([m.v[1]])
        )


@unittest.skipUnless(networkx_available, "networkx is not available.")
class TestIndexedBlock(unittest.TestCase):
    def test_block_data_obj(self):
        m = pyo.ConcreteModel()
        m.block = pyo.Block([1, 2, 3])
        m.block[1].subblock = make_degenerate_solid_phase_model()
        igraph = IncidenceGraphInterface(m.block[1])
        var_dmp, con_dmp = igraph.dulmage_mendelsohn()
        self.assertEqual(len(var_dmp.unmatched), 1)
        self.assertEqual(len(con_dmp.unmatched), 1)

        msg = "Unsupported type.*BlockData"
        with self.assertRaisesRegex(TypeError, msg):
            igraph = IncidenceGraphInterface(m.block)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import sys
import logging

from pyomo.common.deprecation import RenamedClass
from pyomo.common.log import is_debug_set
from pyomo.common.timing import ConstructionTimer

from pyomo.core.base.misc import apply_indexed_rule
from pyomo.core.base.component import ActiveComponentData, ModelComponentFactory
from pyomo.core.base.global_set import UnindexedComponent_index
from pyomo.core.base.indexed_component import (
    ActiveIndexedComponent,
    UnindexedComponent_set,
)
from pyomo.core.base.set_types import PositiveIntegers

logger = logging.getLogger('pyomo.core')


class SOSConstraintData(ActiveComponentData):
    """
    This class defines the data for a single special ordered set.

    Constructor arguments:
        owner           The Constraint object that owns this data.

    Public class attributes:
        active          A boolean that is true if this objective is active in the model.
        component       The constraint component.

    Private class attributes:
        _variables       SOS variables.
        _weights         SOS variable weights.
        _level           SOS level (Positive Integer)
    """

    __slots__ = ('_variables', '_weights', '_level')

    def __init__(self, owner):
        """Constructor"""
        self._level = None
        self._variables = []
        self._weights = []
        ActiveComponentData.__init__(self, owner)

    def num_variables(self):
        return len(self._variables)

    def items(self):
        return zip(self._variables, self._weights)

    @property
    def level(self):
        """
        Return the SOS level
        """
        return self._level

    @level.setter
    def level(self, level):
        if level not in PositiveIntegers:
            raise ValueError("SOS Constraint level must be a positive integer")
        self._level = level

    @property
    def variables(self):
        """
        Return the variable list for the SOS constraint
        """
        return self._variables

    def get_variables(self):
        for val in self._variables:
            yield val

    def get_items(self):
        assert len(self._variables) == len(self._weights)
        for v, w in zip(self._variables, self._weights):
            yield v, w

    def set_items(self, variables, weights):
        self._variables = []
        self._weights = []
        for v, w in zip(variables, weights):
            self._variables.append(v)
            if w < 0.0:
                raise ValueError(
                    "Cannot set negative weight %f for variable %s" % (w, v.name)
                )
            self._weights.append(w)


class _SOSConstraintData(metaclass=RenamedClass):
    __renamed__new_class__ = SOSConstraintData
    __renamed__version__ = '6.7.2'


@ModelComponentFactory.register("SOS constraint expressions.")
class SOSConstraint(ActiveIndexedComponent):
    """
    Implements constraints for special ordered sets (SOS).

    Parameters
    ----------
    sos : int
        The type of SOS.
    var : pyomo.environ.Var
        The group of variables from which the SOS(s) will be created.
    index : pyomo.environ.Set, list or dict, optional
        A data structure with the indexes for the variables that are to be
        members of the SOS(s). The indexes can be provided as a pyomo Set:
        either indexed, if the SOS is indexed; or non-indexed, otherwise.
        Alternatively, the indexes can be provided as a list, for a non-indexed
        SOS, or as a dict, for indexed SOS(s).
    weights : pyomo.environ.Param or dict, optional
        A data structure with the weights for each member of the SOS(s). These
        can be provided as pyomo Param or as a dict. If not provided, the
        weights will be determined automatically using the var index set.
    rule : optional
        A method returning a 2-tuple with lists of variables and the respective
        weights in the same order, or a list of variables whose weights are
        then determined from their position within the list or, alternatively,
        pyomo.environ.Constraint.Skip if the constraint should be not be
        included in the model/instance. This parameter cannot be used in
        combination with var, index or weights.

    Examples
    -------

    1 - An SOS of type **N** made up of all members of a pyomo Var component:

    >>> # import pyomo
    >>> import pyomo.environ as pyo
    >>> # declare the model
    >>> model = pyo.AbstractModel()
    >>> # define the SOS type
    >>> N = 1 # 2, 3, ...
    >>> # the set that indexes the variables
    >>> model.A = pyo.Set()
    >>> # the variables under consideration
    >>> model.x = pyo.Var(model.A)
    >>> # the sos constraint
    >>> model.mysos = pyo.SOSConstraint(var=model.x, sos=N)

    2 - An SOS of type **N** made up of all members of a pyomo Var component,
    each with a specific weight:

    >>> # declare the model
    >>> model = pyo.AbstractModel()
    >>> # define the SOS type
    >>> N = 1 # 2, 3, ...
    >>> # the set that indexes the variables
    >>> model.A = pyo.Set()
    >>> # the variables under consideration
    >>> model.x = pyo.Var(model.A)
    >>> # the weights for each variable used in the sos constraints
    >>> model.mysosweights = pyo.Param(model.A)
    >>> # the sos constraint
    >>> model.mysos = pyo.SOSConstraint(
    ...     var=model.x,
    ...     sos=N,
    ...     weights=model.mysosweights
    ...     )

    3 - An SOS of type **N** made up of selected members of a Var component:

    >>> # declare the model
    >>> model = pyo.AbstractModel()
    >>> # define the SOS type
    >>> N = 1 # 2, 3, ...
    >>> # the set that indexes the variables
    >>> model.A = pyo.Set()
    >>> # the variables under consideration
    >>> model.x = pyo.Var(model.A)
    >>> # the set that indexes the variables actually used in the constraint
    >>> model.B = pyo.Set(within=model.A)
    >>> # the sos constraint
    >>> model.mysos = pyo.SOSConstraint(var=model.x, sos=N, index=model.B)

    4 - An SOS of type **N** made up of selected members of a Var component,
    each with a specific weight:

    >>> # declare the model
    >>> model = pyo.AbstractModel()
    >>> # define the SOS type
    >>> N = 1 # 2, 3, ...
    >>> # the set that indexes the variables
    >>> model.A = pyo.Set()
    >>> # the variables under consideration
    >>> model.x = pyo.Var(model.A)
    >>> # the set that indexes the variables actually used in the constraint
    >>> model.B = pyo.Set(within=model.A)
    >>> # the weights for each variable used in the sos constraints
    >>> model.mysosweights = pyo.Param(model.B)
    >>> # the sos constraint
    >>> model.mysos = pyo.SOSConstraint(
    ...     var=model.x,
    ...     sos=N,
    ...     index=model.B,
    ...     weights=model.mysosweights
    ...     )

    5 - A set of SOS(s) of type **N** made up of members of a pyomo Var
    component:

    >>> # declare the model
    >>> model = pyo.AbstractModel()
    >>> # define the SOS type
    >>> N = 1 # 2, 3, ...
    >>> # the set that indexes the variables
    >>> model.A = pyo.Set()
    >>> # the variables under consideration
    >>> model.x = pyo.Var(model.A)
    >>> # the set indexing the sos constraints
    >>> model.B = pyo.Set()
    >>> # the sets containing the variable indexes for each constraint
    >>> model.mysosvarindexset = pyo.Set(model.B)
    >>> # the sos constraints
    >>> model.mysos = pyo.SOSConstraint(
    ...     model.B,
    ...     var=model.x,
    ...     sos=N,
    ...     index=model.mysosvarindexset
    ...     )

    6 - A set of SOS(s) of type **N** made up of members of a pyomo Var
    component, each with a specific weight:

    >>> # declare the model
    >>> model = pyo.AbstractModel()
    >>> # define the SOS type
    >>> N = 1 # 2, 3, ...
    >>> # the set that indexes the variables
    >>> model.A = pyo.Set()
    >>> # the variables under consideration
    >>> model.x = pyo.Var(model.A)
    >>> # the set indexing the sos constraints
    >>> model.B = pyo.Set()
    >>> # the sets containing the variable indexes for each constraint
    >>> model.mysosvarindexset = pyo.Set(model.B)
    >>> # the set that indexes the variables used in the sos constraints
    >>> model.C = pyo.Set(within=model.A)
    >>> # the weights for each variable used in the sos constraints
    >>> model.mysosweights = pyo.Param(model.C)
    >>> # the sos constraints
    >>> model.mysos = pyo.SOSConstraint(
    ...     model.B,
    ...     var=model.x,
    ...     sos=N,
    ...     index=model.mysosvarindexset,
    ...     weights=model.mysosweights,
    ...     )

    7 - A simple SOS of type **N** created using the rule parameter:

    >>> # declare the model
    >>> model = pyo.AbstractModel()
    >>> # define the SOS type
    >>> N = 1 # 2, 3, ...
    >>> # the set that indexes the variables
    >>> model.A = pyo.Set()
    >>> # the variables under consideration
    >>> model.x = pyo.Var(model.A, domain=pyo.NonNegativeReals)
    >>> # the rule method creating the constraint
    >>> def rule_mysos(m):
    ...     var_list = [m.x[a] for a in m.x]
    ...     weight_list = [i+1 for i in range(len(var_list))]
    ...     return (var_list, weight_list)
    >>> # the sos constraint(s)
    >>> model.mysos = pyo.SOSConstraint(rule=rule_mysos, sos=N)

    8 - A simple SOS of type **N** created using the rule parameter, in which
    the weights are determined automatically:

    >>> # declare the model
    >>> model = pyo.AbstractModel()
    >>> # define the SOS type
    >>> N = 1 # 2, 3, ...
    >>> # the set that indexes the variables
    >>> model.A = pyo.Set()
    >>> # the variables under consideration
    >>> model.x = pyo.Var(model.A, domain=pyo.NonNegativeReals)
    >>> # the rule method creating the constraint
    >>> def rule_mysos(m):
    ...     return [m.x[a] for a in m.x]
    >>> # the sos constraint(s)
    >>> model.mysos = pyo.SOSConstraint(rule=rule_mysos, sos=N)

    9 - A set of SOS(s) of type **N** involving members of distinct pyomo Var
    components, each with a specific weight. This requires the rule parameter:

    >>> # declare the model
    >>> model = pyo.AbstractModel()
    >>> # define the SOS type
    >>> N = 1 # 2, 3, ...
    >>> # the set that indexes the x variables
    >>> model.A = pyo.Set()
    >>> # the set that indexes the y variables
    >>> model.B = pyo.Set()
    >>> # the set that indexes the SOS constraints
    >>> model.C = pyo.Set()
    >>> # the x variables, which will be used in the constraints
    >>> model.x = pyo.Var(model.A, domain=pyo.NonNegativeReals)
    >>> # the y variables, which will be used in the constraints
    >>> model.y = pyo.Var(model.B, domain=pyo.NonNegativeReals)
    >>> # the x variable indices for each constraint
    >>> model.mysosindex_x = pyo.Set(model.C)
    >>> # the y variable indices for each constraint
    >>> model.mysosindex_y = pyo.Set(model.C)
    >>> # the weights for the x variable indices
    >>> model.mysosweights_x = pyo.Param(model.A)
    >>> # the weights for the y variable indices
    >>> model.mysosweights_y = pyo.Param(model.B)
    >>> # the rule method with which each constraint c is built
    >>> def rule_mysos(m, c):
    ...     var_list = [m.x[a] for a in m.mysosindex_x[c]]
    ...     var_list.extend([m.y[b] for b in m.mysosindex_y[c]])
    ...     weight_list = [m.mysosweights_x[a] for a in m.mysosindex_x[c]]
    ...     weight_list.extend([m.mysosweights_y[b] for b in m.mysosindex_y[c]])
    ...     return (var_list, weight_list)
    >>> # the sos constraint(s)
    >>> model.mysos = pyo.SOSConstraint(
    ...     model.C,
    ...     rule=rule_mysos,
    ...     sos=N
    ...     )

    """

    Skip = (1000,)

    def __new__(cls, *args, **kwds):
        if cls != SOSConstraint:
            return super(SOSConstraint, cls).__new__(cls)
        if not args or (args[0] is UnindexedComponent_set and len(args) == 1):
            return ScalarSOSConstraint.__new__(ScalarSOSConstraint)
        else:
            return IndexedSOSConstraint.__new__(IndexedSOSConstraint)

    def __init__(self, *args, **kwargs):
        """
        Constructor
        """
        #
        # The 'initialize' or 'rule' argument
        #
        initialize = kwargs.pop('initialize', None)
        initialize = kwargs.pop('rule', initialize)
        if not initialize is None:
            if 'var' in kwargs:
                raise TypeError(
                    "Cannot specify the 'var' argument with the 'rule' or 'initialize' argument"
                )
            if 'index' in kwargs:
                raise TypeError(
                    "Cannot specify the 'index' argument with the 'rule' or 'initialize' argument"
                )
            if 'weights' in kwargs:
                raise TypeError(
                    "Cannot specify the 'weights' argument with the 'rule' or 'initialize' argument"
                )
        #
        # The 'var' argument
        #
        sosVars = kwargs.pop('var', None)
        if sosVars is None and initialize is None:
            raise TypeError(
                "SOSConstraint() requires either the 'var' or 'initialize' arguments"
            )
        #
        # The 'weights' argument
        #
        sosWeights = kwargs.pop('weights', None)
        #
        # The 'index' argument
        #
        sosSet = kwargs.pop('index', None)
        #
        # The 'sos' or 'level' argument
        #
        if 'sos' in kwargs and 'level' in kwargs:
            raise TypeError(
                "Specify only one of 'sos' and 'level' -- "
                "they are equivalent keyword arguments"
            )
        sosLevel = kwargs.pop('sos', None)
        sosLevel = kwargs.pop('level', sosLevel)
        if sosLevel is None:
            raise TypeError(
                "SOSConstraint() requires that either the "
                "'sos' or 'level' keyword arguments be set to indicate "
                "the type of SOS."
            )
        #
        # Set attributes
        #
        self._sosVars = sosVars
        self._sosWeights = sosWeights
        self._sosSet = sosSet
        self._sosLevel = sosLevel
        self._rule = initialize
        #
        # Construct the base class
        #
        kwargs.setdefault('ctype', SOSConstraint)
        ActiveIndexedComponent.__init__(self, *args, **kwargs)

    def construct(self, data=None):
        """
        Construct this component
        """
        assert data is None  # because I don't know why it's an argument
        generate_debug_messages = is_debug_set(logger)
        if self._constructed is True:  # pragma:nocover
            return

        if generate_debug_messages:  # pragma:nocover
            logger.debug("Constructing SOSConstraint %s", self.name)
        timer = ConstructionTimer(self)
        self._constructed = True

        if self._rule is None:
            if self._sosSet is None and self.is_indexed():
                if generate_debug_messages:  # pragma:nocover
                    logger.debug(
                        "  Cannot construct "
                        + self.name
                        + ".  No rule is defined and no SOS sets are defined."
                    )
            else:
                if not self.is_indexed():
                    if self._sosSet is None:
                        if getattr(
                            self._sosVars.index_set(), 'isordered', lambda *x: False
                        )():
                            _sosSet = {None: list(self._sosVars.index_set())}
                        else:
                            _sosSet = {None: set(self._sosVars.index_set())}
                    else:
                        _sosSet = {None: self._sosSet}
                else:
                    _sosSet = self._sosSet

                for index, sosSet in _sosSet.items():
                    if generate_debug_messages:  # pragma:nocover
                        logger.debug(
                            "  Constructing " + self.name + " index " + str(index)
                        )

                    if self._sosLevel == 2:
                        #
                        # Check that the sets are ordered.
                        #
                        ordered = False
                        if (
                            type(sosSet) is list
                            or sosSet is UnindexedComponent_set
                            or len(sosSet) == 1
                        ):
                            ordered = True
                        if hasattr(sosSet, 'isordered') and sosSet.isordered():
                            ordered = True
                        if not ordered:
                            raise ValueError(
                                "Cannot define a SOS over an unordered index."
                            )

                    variables = [self._sosVars[idx] for idx in sosSet]
                    if self._sosWeights is not None:
                        weights = [self._sosWeights[idx] for idx in sosSet]
                    else:
                        weights = None

                    self.add(index, variables, weights)
        else:
            _self_rule = self._rule
            _self_parent = self._parent()
            for index in self._index_set:
                try:
                    tmp = apply_indexed_rule(self, _self_rule, _self_parent, index)
                except Exception:
                    err = sys.exc_info()[1]
                    logger.error(
                        "Rule failed when generating expression for "
                        "sos constraint %s with index %s:\n%s: %s"
                        % (self.name, str(index), type(err).__name__, err)
                    )
                    raise
                if tmp is None:
                    raise ValueError(
                        "SOSConstraint rule returned None instead of SOSConstraint.Skip for index %s"
                        % str(index)
                    )
                if type(tmp) is tuple:
                    if tmp is SOSConstraint.Skip:
                        continue
                    # tmp is a tuple of variables, weights
                    self.add(index, tmp[0], tmp[1])
                else:
                    # tmp is a list of variables
                    self.add(index, tmp)
        timer.report()

    def add(self, index, variables, weights=None):
        """
        Add a component data for the specified index.
        """
        if index is None:
            # because ScalarSOSConstraint already makes an SOSConstraintData instance
            soscondata = self
        else:
            soscondata = SOSConstraintData(self)
        self._data[index] = soscondata
        soscondata._index = index

        soscondata.level = self._sosLevel

        if weights is None:
            soscondata.set_items(variables, list(range(1, len(variables) + 1)))
        else:
            soscondata.set_items(variables, weights)

    def _pprint(self):
        """Print component information."""
        headers = [
            ("Size", len(self)),
            ("Index", self._index_set if self.is_indexed() else None),
        ]
        return (
            headers,
            self.items,
            ("Type", "Weight", "Variable"),
            lambda k, v: (
                ("" if i else v.level, w, var)
                for i, (var, w) in enumerate(v.get_items())
            ),
        )


class ScalarSOSConstraint(SOSConstraint, SOSConstraintData):
    def __init__(self, *args, **kwd):
        SOSConstraintData.__init__(self, self)
        SOSConstraint.__init__(self, *args, **kwd)
        self._index = UnindexedComponent_index


class SimpleSOSConstraint(metaclass=RenamedClass):
    __renamed__new_class__ = ScalarSOSConstraint
    __renamed__version__ = '6.0'


class IndexedSOSConstraint(SOSConstraint):
    def __init__(self, *args, **kwds):
        super(IndexedSOSConstraint, self).__init__(*args, **kwds)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.kernel as pmo
from pyomo.core import (
    ConcreteModel,
    Var,
    Objective,
    Constraint,
    RangeSet,
    ConstraintList,
)
from pyomo.solvers.tests.models.base import _BaseTestModel, register_model


@register_model
class LP_trivial_constraints(_BaseTestModel):
    """
    A continuous linear model with trivial constraints
    """

    description = "LP_trivial_constraints"
    capabilities = set(['linear'])
    test_pickling = False

    def __init__(self):
        _BaseTestModel.__init__(self)
        self.add_results(self.description + ".json")

    def _generate_model(self):
        self.model = None
        self.model = ConcreteModel()
        model = self.model
        model._name = self.description

        model.x = Var(domain=RangeSet(float('-inf'), None, 0))
        model.y = Var(bounds=(None, float('inf')))
        model.obj = Objective(expr=model.x - model.y)
        model.c = ConstraintList()
        model.c.add(model.x >= -2)
        model.c.add(model.y <= 3)
        cdata = model.c.add((0, 1, 3))
        assert cdata.lower == 0
        assert cdata.upper == 3
        assert cdata.body() == 1
        assert not cdata.equality
        cdata = model.c.add((0, 2, 3))
        assert cdata.lower == 0
        assert cdata.upper == 3
        assert cdata.body() == 2
        assert not cdata.equality
        cdata = model.c.add((0, 1, None))
        assert cdata.lower is None
        assert cdata.upper == 1
        assert cdata.body() == 0
        assert not cdata.equality
        cdata = model.c.add((None, 0, 1))
        assert cdata.lower is None
        assert cdata.upper == 1
        assert cdata.body() == 0
        assert not cdata.equality
        cdata = model.c.add((1, 1))
        assert cdata.lower == 1
        assert cdata.upper == 1
        assert cdata.body() == 1
        assert cdata.equality
        model.d = Constraint(rule=lambda m: (float('-inf'), m.x, float('inf')))
        assert not model.d.equality

    def warmstart_model(self):
        assert self.model is not None
        pass

    def post_solve_test_validation(self, tester, results):
        symbol_map = results._smap
        assert not symbol_map is None
        if tester is None:
            for i in self.model.c:
                assert id(self.model.c[i]) in symbol_map.byObject
            assert id(self.model.d[i]) not in symbol_map.byObject
        else:
            for i in self.model.c:
                tester.assertIn(id(self.model.c[i]), symbol_map.byObject)
            tester.assertNotIn(id(self.model.d), symbol_map.byObject)


@register_model
class LP_trivial_constraints_kernel(LP_trivial_constraints):
    def _generate_model(self):
        self.model = None
        self.model = pmo.block()
        model = self.model
        model._name = self.description

        model.x = pmo.variable(domain=RangeSet(float('-inf'), None, 0))
        model.y = pmo.variable(ub=float('inf'))
        model.obj = pmo.objective(model.x - model.y)
        model.c = pmo.constraint_dict()
        model.c[1] = pmo.constraint(model.x >= -2)
        model.c[2] = pmo.constraint(model.y <= 3)
        cdata = model.c[3] = pmo.constraint((0, 1, 3))
        assert cdata.lb == 0
        assert cdata.ub == 3
        assert cdata.body() == 1
        assert not cdata.equality
        cdata = model.c[4] = pmo.constraint((0, 2, 3))
        assert cdata.lb == 0
        assert cdata.ub == 3
        assert cdata.body() == 2
        assert not cdata.equality
        cdata = model.c[5] = pmo.constraint((0, 1, None))
        assert cdata.lb is None
        assert cdata.ub == 1
        assert cdata.body() == 0
        assert not cdata.equality
        cdata = model.c[6] = pmo.constraint((None, 0, 1))
        assert cdata.lb is None
        assert cdata.ub == 1
        assert cdata.body() == 0
        assert not cdata.equality
        cdata = model.c[7] = pmo.constraint((1, 1))
        assert cdata.lb == 1
        assert cdata.ub == 1
        assert cdata.body() == 1
        assert cdata.equality

    def post_solve_test_validation(self, tester, results):
        symbol_map = results.Solution(0).symbol_map
        assert not symbol_map is None
        if tester is None:
            for i in self.model.c:
                assert id(self.model.c[i]) in symbol_map.byObject
        else:
            tester.assertNotEqual(symbol_map, None)
            for i in self.model.c:
                tester.assertTrue(id(self.model.c[i]) in symbol_map.byObject)



# -*- coding: utf-8 -*-
# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

from io import StringIO

import pyomo.common.unittest as unittest

from pyomo.common.log import LoggingIntercept
from pyomo.environ import ConcreteModel, Constraint, Var, inequality
from pyomo.util.blockutil import log_model_constraints


class TestBlockutil(unittest.TestCase):
    """Tests block utilities."""

    def test_log_model_constraints(self):
        m = ConcreteModel()
        m.x = Var(initialize=1)
        m.c1 = Constraint(expr=m.x >= 2)
        m.c2 = Constraint(expr=m.x == 4)
        m.c3 = Constraint(expr=m.x <= 0)
        m.y = Var(bounds=(0, 2), initialize=1.9999999)
        m.c4 = Constraint(expr=m.y >= m.y.value)
        m.z = Var(bounds=(0, 6))
        m.c5 = Constraint(expr=inequality(5, m.z, 10), doc="Range infeasible")
        m.c6 = Constraint(expr=m.x + m.y <= 6, doc="Feasible")
        m.c7 = Constraint(expr=m.z == 6, doc="Equality infeasible")
        m.c8 = Constraint(expr=inequality(3, m.x, 6), doc="Range lb infeasible")
        m.c9 = Constraint(expr=inequality(0, m.x, 0.5), doc="Range ub infeasible")
        m.c10 = Constraint(expr=m.y >= 3, doc="Inactive")
        m.c10.deactivate()
        m.c11 = Constraint(expr=m.y <= m.y.value)
        m.yy = Var(bounds=(0, 1), initialize=1e-7, doc="Close to lower bound")
        m.y3 = Var(bounds=(0, 1e-7), initialize=0, doc="Bounds too close")
        m.y4 = Var(bounds=(0, 1), initialize=2, doc="Fixed out of bounds.")
        m.y4.fix()

        output = StringIO()
        with LoggingIntercept(output, 'pyomo.util', logging.INFO):
            log_model_constraints(m)
        expected_output = [
            "c1 active",
            "c2 active",
            "c3 active",
            "c4 active",
            "c5 active",
            "c6 active",
            "c7 active",
            "c8 active",
            "c9 active",
            "c11 active",
        ]
        self.assertEqual(expected_output, output.getvalue().splitlines())



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import math
import pyomo.common.unittest as unittest

from pyomo.environ import (
    exp,
    log,
    log10,
    sin,
    cos,
    tan,
    asin,
    acos,
    atan,
    sqrt,
    inequality,
    Expr_if,
    Any,
    ConcreteModel,
    Expression,
    Param,
    Var,
)

from pyomo.common.errors import DeveloperError
from pyomo.common.log import LoggingIntercept
from pyomo.contrib.fbbt.expression_bounds_walker import ExpressionBoundsVisitor, inf
from pyomo.contrib.fbbt.interval import _true, _false
from pyomo.core.expr import ExpressionBase, NumericExpression, BooleanExpression


class TestExpressionBoundsWalker(unittest.TestCase):
    def make_model(self):
        m = ConcreteModel()
        m.x = Var(bounds=(-2, 4))
        m.y = Var(bounds=(3, 5))
        m.z = Var(bounds=(0.5, 0.75))
        return m

    def test_sum_bounds(self):
        m = self.make_model()
        visitor = ExpressionBoundsVisitor()
        lb, ub = visitor.walk_expression(m.x + m.y)
        self.assertEqual(lb, 1)
        self.assertEqual(ub, 9)

        self.assertEqual(len(visitor.leaf_bounds), 2)
        self.assertIn(m.x, visitor.leaf_bounds)
        self.assertIn(m.y, visitor.leaf_bounds)
        self.assertEqual(visitor.leaf_bounds[m.x], (-2, 4))
        self.assertEqual(visitor.leaf_bounds[m.y], (3, 5))

    def test_fixed_var(self):
        m = self.make_model()
        m.x.fix(3)

        visitor = ExpressionBoundsVisitor()
        lb, ub = visitor.walk_expression(m.x + m.y)
        self.assertEqual(lb, 1)
        self.assertEqual(ub, 9)

        self.assertEqual(len(visitor.leaf_bounds), 2)
        self.assertIn(m.x, visitor.leaf_bounds)
        self.assertIn(m.y, visitor.leaf_bounds)
        self.assertEqual(visitor.leaf_bounds[m.x], (-2, 4))
        self.assertEqual(visitor.leaf_bounds[m.y], (3, 5))

    def test_fixed_var_value_used_for_bounds(self):
        m = self.make_model()
        m.x.fix(3)

        visitor = ExpressionBoundsVisitor(use_fixed_var_values_as_bounds=True)
        lb, ub = visitor.walk_expression(m.x + m.y)
        self.assertEqual(lb, 6)
        self.assertEqual(ub, 8)

        self.assertEqual(len(visitor.leaf_bounds), 2)
        self.assertIn(m.x, visitor.leaf_bounds)
        self.assertIn(m.y, visitor.leaf_bounds)
        self.assertEqual(visitor.leaf_bounds[m.x], (3, 3))
        self.assertEqual(visitor.leaf_bounds[m.y], (3, 5))

    def test_product_bounds(self):
        m = self.make_model()
        visitor = ExpressionBoundsVisitor()
        lb, ub = visitor.walk_expression(m.x * m.y)
        self.assertEqual(lb, -10)
        self.assertEqual(ub, 20)

    def test_division_bounds(self):
        m = self.make_model()
        visitor = ExpressionBoundsVisitor()
        lb, ub = visitor.walk_expression(m.x / m.y)
        self.assertAlmostEqual(lb, -2 / 3)
        self.assertAlmostEqual(ub, 4 / 3)

    def test_power_bounds(self):
        m = self.make_model()
        visitor = ExpressionBoundsVisitor()
        lb, ub = visitor.walk_expression(m.y**m.x)
        self.assertEqual(lb, 5 ** (-2))
        self.assertEqual(ub, 5**4)

    def test_sums_of_squares_bounds(self):
        m = ConcreteModel()
        m.x = Var([1, 2], bounds=(-2, 6))
        visitor = ExpressionBoundsVisitor()
        lb, ub = visitor.walk_expression(m.x[1] * m.x[1] + m.x[2] * m.x[2])
        self.assertEqual(lb, 0)
        self.assertEqual(ub, 72)

    def test_negation_bounds(self):
        m = self.make_model()
        visitor = ExpressionBoundsVisitor()
        lb, ub = visitor.walk_expression(-(m.y + 3 * m.x))
        self.assertEqual(lb, -17)
        self.assertEqual(ub, 3)

    def test_exp_bounds(self):
        m = self.make_model()
        visitor = ExpressionBoundsVisitor()
        lb, ub = visitor.walk_expression(exp(m.y))
        self.assertAlmostEqual(lb, math.e**3)
        self.assertAlmostEqual(ub, math.e**5)

    def test_log_bounds(self):
        m = self.make_model()
        visitor = ExpressionBoundsVisitor()
        lb, ub = visitor.walk_expression(log(m.y))
        self.assertAlmostEqual(lb, log(3))
        self.assertAlmostEqual(ub, log(5))

    def test_log10_bounds(self):
        m = self.make_model()
        visitor = ExpressionBoundsVisitor()
        lb, ub = visitor.walk_expression(log10(m.y))
        self.assertAlmostEqual(lb, log10(3))
        self.assertAlmostEqual(ub, log10(5))

    def test_sin_bounds(self):
        m = self.make_model()
        visitor = ExpressionBoundsVisitor()
        lb, ub = visitor.walk_expression(sin(m.y))
        self.assertAlmostEqual(lb, -1)  # reaches -1 at 3*pi/2 \approx 4.712
        self.assertAlmostEqual(ub, sin(3))  # it's positive here

    def test_cos_bounds(self):
        m = self.make_model()
        visitor = ExpressionBoundsVisitor()
        lb, ub = visitor.walk_expression(cos(m.y))
        self.assertAlmostEqual(lb, -1)  # reaches -1 at pi
        self.assertAlmostEqual(ub, cos(5))  # it's positive here

    def test_tan_bounds(self):
        m = self.make_model()
        visitor = ExpressionBoundsVisitor()
        lb, ub = visitor.walk_expression(tan(m.y))
        self.assertEqual(lb, -float('inf'))
        self.assertEqual(ub, float('inf'))

    def test_asin_bounds(self):
        m = self.make_model()
        visitor = ExpressionBoundsVisitor()
        lb, ub = visitor.walk_expression(asin(m.z))
        self.assertAlmostEqual(lb, asin(0.5))
        self.assertAlmostEqual(ub, asin(0.75))

    def test_acos_bounds(self):
        m = self.make_model()
        visitor = ExpressionBoundsVisitor()
        lb, ub = visitor.walk_expression(acos(m.z))
        self.assertAlmostEqual(lb, acos(0.75))
        self.assertAlmostEqual(ub, acos(0.5))

    def test_atan_bounds(self):
        m = self.make_model()
        visitor = ExpressionBoundsVisitor()
        lb, ub = visitor.walk_expression(atan(m.z))
        self.assertAlmostEqual(lb, atan(0.5))
        self.assertAlmostEqual(ub, atan(0.75))

    def test_sqrt_bounds(self):
        m = self.make_model()
        visitor = ExpressionBoundsVisitor()
        lb, ub = visitor.walk_expression(sqrt(m.y))
        self.assertAlmostEqual(lb, sqrt(3))
        self.assertAlmostEqual(ub, sqrt(5))

    def test_abs_bounds(self):
        m = self.make_model()
        visitor = ExpressionBoundsVisitor()
        lb, ub = visitor.walk_expression(abs(m.x))
        self.assertEqual(lb, 0)
        self.assertEqual(ub, 4)

    def test_leaf_bounds_cached(self):
        m = self.make_model()
        visitor = ExpressionBoundsVisitor()
        lb, ub = visitor.walk_expression(m.x - m.y)
        self.assertEqual(lb, -7)
        self.assertEqual(ub, 1)

        self.assertIn(m.x, visitor.leaf_bounds)
        self.assertEqual(visitor.leaf_bounds[m.x], m.x.bounds)
        self.assertIn(m.y, visitor.leaf_bounds)
        self.assertEqual(visitor.leaf_bounds[m.y], m.y.bounds)

        # This should exercise the code that uses the cache.
        lb, ub = visitor.walk_expression(m.x**2 + 3)
        self.assertEqual(lb, 3)
        self.assertEqual(ub, 19)

    def test_var_fixed_to_None(self):
        m = self.make_model()
        m.x.fix(None)

        visitor = ExpressionBoundsVisitor(use_fixed_var_values_as_bounds=True)
        with self.assertRaisesRegex(
            ValueError,
            "Var 'x' is fixed to None. This value cannot be "
            "used to calculate bounds.",
        ):
            lb, ub = visitor.walk_expression(m.x - m.y)

    def test_var_with_no_lb(self):
        m = self.make_model()
        m.x.setlb(None)

        visitor = ExpressionBoundsVisitor()
        lb, ub = visitor.walk_expression(m.x - m.y)
        self.assertEqual(lb, -float('inf'))
        self.assertEqual(ub, 1)

    def test_var_with_no_ub(self):
        m = self.make_model()
        m.y.setub(None)

        visitor = ExpressionBoundsVisitor()
        lb, ub = visitor.walk_expression(m.x - m.y)
        self.assertEqual(lb, -float('inf'))
        self.assertEqual(ub, 1)

    def test_param(self):
        m = self.make_model()
        m.p = Param(initialize=6)

        visitor = ExpressionBoundsVisitor()
        lb, ub = visitor.walk_expression(m.p**m.y)
        self.assertEqual(lb, 6**3)
        self.assertEqual(ub, 6**5)

    def test_mutable_param(self):
        m = self.make_model()
        m.p = Param(initialize=6, mutable=True)

        visitor = ExpressionBoundsVisitor()
        lb, ub = visitor.walk_expression(m.p**m.y)
        self.assertEqual(lb, 6**3)
        self.assertEqual(ub, 6**5)

    def test_named_expression(self):
        m = self.make_model()
        m.e = Expression(expr=sqrt(m.x**2 + m.y**2))
        visitor = ExpressionBoundsVisitor()

        lb, ub = visitor.walk_expression(m.e + 4)
        self.assertEqual(lb, 7)
        self.assertAlmostEqual(ub, sqrt(41) + 4)

        self.assertIn(m.e, visitor.leaf_bounds)
        self.assertEqual(visitor.leaf_bounds[m.e][0], 3)
        self.assertAlmostEqual(visitor.leaf_bounds[m.e][1], sqrt(41))

        # exercise the using of the cached bounds
        lb, ub = visitor.walk_expression(m.e)
        self.assertEqual(lb, 3)
        self.assertAlmostEqual(ub, sqrt(41))

    def test_npv_expression(self):
        m = self.make_model()
        m.p = Param(initialize=4, mutable=True)
        visitor = ExpressionBoundsVisitor()
        lb, ub = visitor.walk_expression(1 / m.p)
        self.assertEqual(lb, 0.25)
        self.assertEqual(ub, 0.25)

    def test_invalid_numeric_type(self):
        m = self.make_model()
        m.p = Param(initialize=True, mutable=True, domain=Any)
        visitor = ExpressionBoundsVisitor()
        with self.assertRaisesRegex(
            ValueError,
            r"True \(bool\) is not a valid numeric type. "
            r"Cannot compute bounds on expression.",
        ):
            lb, ub = visitor.walk_expression(m.p + m.y)

        m.p.set_value(None)
        with self.assertRaisesRegex(
            ValueError,
            r"None \(NoneType\) is not a valid numeric type. "
            r"Cannot compute bounds on expression.",
        ):
            lb, ub = visitor.walk_expression(m.p + m.y)

    def test_invalid_string(self):
        m = self.make_model()
        m.p = Param(initialize='True', domain=Any)
        visitor = ExpressionBoundsVisitor()
        with self.assertRaisesRegex(
            ValueError,
            r"'True' \(str\) is not a valid numeric type. "
            r"Cannot compute bounds on expression.",
        ):
            lb, ub = visitor.walk_expression(m.p + m.y)

    def test_invalid_complex(self):
        m = self.make_model()
        m.p = Param(initialize=complex(4, 5), domain=Any)
        visitor = ExpressionBoundsVisitor()
        with self.assertRaisesRegex(
            ValueError,
            r"Cannot compute bounds on expressions containing "
            r"complex numbers. Encountered when processing \(4\+5j\)",
        ):
            lb, ub = visitor.walk_expression(m.p + m.y)

    def test_inequality(self):
        m = self.make_model()
        visitor = ExpressionBoundsVisitor()
        self.assertEqual(visitor.walk_expression(m.z <= m.y), (_true, _true))
        self.assertEqual(visitor.walk_expression(m.y <= m.z), (_false, _false))
        self.assertEqual(visitor.walk_expression(m.y <= m.x), (_false, _true))

    def test_equality(self):
        m = self.make_model()
        m.p = Param(initialize=5)
        visitor = ExpressionBoundsVisitor()
        self.assertEqual(visitor.walk_expression(m.y == m.z), (_false, _false))
        self.assertEqual(visitor.walk_expression(m.y == m.x), (_false, _true))
        self.assertEqual(visitor.walk_expression(m.p == m.p), (_true, _true))

    def test_ranged(self):
        m = self.make_model()
        visitor = ExpressionBoundsVisitor()
        self.assertEqual(
            visitor.walk_expression(inequality(m.z, m.y, 5)), (_true, _true)
        )
        self.assertEqual(
            visitor.walk_expression(inequality(m.y, m.z, m.y)), (_false, _false)
        )
        self.assertEqual(
            visitor.walk_expression(inequality(m.y, m.x, m.y)), (_false, _true)
        )

    def test_expr_if(self):
        m = self.make_model()
        visitor = ExpressionBoundsVisitor()
        self.assertEqual(
            visitor.walk_expression(Expr_if(IF=m.z <= m.y, THEN=m.z, ELSE=m.y)),
            m.z.bounds,
        )
        self.assertEqual(
            visitor.walk_expression(Expr_if(IF=m.z >= m.y, THEN=m.z, ELSE=m.y)),
            m.y.bounds,
        )
        self.assertEqual(
            visitor.walk_expression(Expr_if(IF=m.y <= m.x, THEN=m.y, ELSE=m.x)), (-2, 5)
        )

    def test_unknown_classes(self):
        class UnknownNumeric(NumericExpression):
            pass

        class UnknownLogic(BooleanExpression):
            def nargs(self):
                return 0

        class UnknownOther(ExpressionBase):
            @property
            def args(self):
                return ()

            def nargs(self):
                return 0

        visitor = ExpressionBoundsVisitor()
        with LoggingIntercept() as LOG:
            self.assertEqual(visitor.walk_expression(UnknownNumeric(())), (-inf, inf))
        self.assertEqual(
            LOG.getvalue(),
            "Unexpected expression node type 'UnknownNumeric' found while walking "
            "expression tree; returning (-inf, inf) for the expression bounds.\n",
        )
        with LoggingIntercept() as LOG:
            self.assertEqual(visitor.walk_expression(UnknownLogic(())), (_false, _true))
        self.assertEqual(
            LOG.getvalue(),
            "Unexpected expression node type 'UnknownLogic' found while walking "
            "expression tree; returning (False, True) for the expression bounds.\n",
        )
        with self.assertRaisesRegex(
            DeveloperError, "Unexpected expression node type 'UnknownOther' found"
        ):
            visitor.walk_expression(UnknownOther())



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from math import fabs
from pyomo.environ import value


def check_8PP_solution(self, eight_process, results):
    self.assertTrue(fabs(value(eight_process.profit.expr) - 68) <= 1e-2)
    self.assertTrue(fabs(value(results.problem.upper_bound) - 68) <= 1e-2)

    # Check discrete solution
    # use 2
    self.assertTrue(value(eight_process.use_unit_1or2.disjuncts[1].indicator_var))
    self.assertFalse(value(eight_process.use_unit_1or2.disjuncts[0].indicator_var))
    # use 4
    self.assertTrue(value(eight_process.use_unit_4or5ornot.disjuncts[0].indicator_var))
    self.assertFalse(value(eight_process.use_unit_4or5ornot.disjuncts[1].indicator_var))
    self.assertFalse(value(eight_process.use_unit_4or5ornot.disjuncts[2].indicator_var))
    # use 6
    self.assertTrue(value(eight_process.use_unit_6or7ornot.disjuncts[0].indicator_var))
    self.assertFalse(value(eight_process.use_unit_6or7ornot.disjuncts[1].indicator_var))
    self.assertFalse(value(eight_process.use_unit_6or7ornot.disjuncts[2].indicator_var))
    # use 8
    self.assertTrue(value(eight_process.use_unit_8ornot.disjuncts[0].indicator_var))
    self.assertFalse(value(eight_process.use_unit_8ornot.disjuncts[1].indicator_var))


def check_8PP_logical_solution(self, eight_process, results):
    self.assertTrue(fabs(value(eight_process.profit.expr) - 68) <= 1e-2)
    self.assertTrue(fabs(value(results.problem.upper_bound) - 68) <= 1e-2)

    # Check discrete solution
    # use 2
    self.assertTrue(value(eight_process.use_unit[2].indicator_var))
    self.assertFalse(value(eight_process.no_unit[2].indicator_var))
    # use 4
    self.assertTrue(value(eight_process.use_unit[4].indicator_var))
    self.assertFalse(value(eight_process.no_unit[4].indicator_var))
    # use 6
    self.assertTrue(value(eight_process.use_unit[6].indicator_var))
    self.assertFalse(value(eight_process.no_unit[6].indicator_var))
    # use 8
    self.assertTrue(value(eight_process.use_unit[8].indicator_var))
    self.assertFalse(value(eight_process.no_unit[8].indicator_var))

    self.assertFalse(value(eight_process.use_unit[1].indicator_var))
    self.assertTrue(value(eight_process.no_unit[1].indicator_var))
    self.assertFalse(value(eight_process.use_unit[3].indicator_var))
    self.assertTrue(value(eight_process.no_unit[3].indicator_var))
    self.assertFalse(value(eight_process.use_unit[5].indicator_var))
    self.assertTrue(value(eight_process.no_unit[5].indicator_var))
    self.assertFalse(value(eight_process.use_unit[7].indicator_var))
    self.assertTrue(value(eight_process.no_unit[7].indicator_var))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
"""
The cyipopt_solver module includes two solvers that call CyIpopt. One,
CyIpoptSolver, is a solver that operates on a CyIpoptProblemInterface
(such as CyIpoptNLP). The other, PyomoCyIpoptSolver, operates directly on a
Pyomo model.

"""

import io
import sys
import logging
import os
import abc

from pyomo.common.deprecation import relocated_module_attribute
from pyomo.common.dependencies import attempt_import, numpy as np, numpy_available
from pyomo.common.tee import capture_output
from pyomo.common.modeling import unique_component_name
from pyomo.core.base.objective import Objective

# Because pynumero.interfaces requires numpy, we will leverage deferred
# imports here so that the solver can be registered even when numpy is
# not available.
pyomo_nlp = attempt_import("pyomo.contrib.pynumero.interfaces.pyomo_nlp")[0]
pyomo_grey_box = attempt_import("pyomo.contrib.pynumero.interfaces.pyomo_grey_box_nlp")[
    0
]
egb = attempt_import("pyomo.contrib.pynumero.interfaces.external_grey_box")[0]

# Defer this import so that importing this module (PyomoCyIpoptSolver in
# particular) does not rely on an attempted cyipopt import.
cyipopt_interface, _ = attempt_import(
    "pyomo.contrib.pynumero.interfaces.cyipopt_interface"
)

# These attributes should no longer be imported from this module. These
# deprecation paths provide a deferred import to these attributes so (a) they
# can still be used until these paths are removed, and (b) the imports are not
# triggered when this module is imported.
relocated_module_attribute(
    "cyipopt_available",
    "pyomo.contrib.pynumero.interfaces.cyipopt_interface.cyipopt_available",
    "6.6.0",
)
relocated_module_attribute(
    "CyIpoptProblemInterface",
    "pyomo.contrib.pynumero.interfaces.cyipopt_interface.CyIpoptProblemInterface",
    "6.6.0",
)
relocated_module_attribute(
    "CyIpoptNLP",
    "pyomo.contrib.pynumero.interfaces.cyipopt_interface.CyIpoptNLP",
    "6.6.0",
)

from pyomo.common.config import ConfigBlock, ConfigValue
from pyomo.common.timing import TicTocTimer
from pyomo.core.base import Block, Objective, minimize
from pyomo.opt import SolverStatus, SolverResults, TerminationCondition
from pyomo.opt.results.solution import Solution

logger = logging.getLogger(__name__)

# This maps the cyipopt STATUS_MESSAGES back to string representations
# of the Ipopt ApplicationReturnStatus enum
_cyipopt_status_enum = [
    "Solve_Succeeded",
    (
        b"Algorithm terminated successfully at a locally "
        b"optimal point, satisfying the convergence tolerances "
        b"(can be specified by options)."
    ),
    "Solved_To_Acceptable_Level",
    (
        b"Algorithm stopped at a point that was "
        b'converged, not to "desired" tolerances, '
        b'but to "acceptable" tolerances (see the '
        b"acceptable-... options)."
    ),
    "Infeasible_Problem_Detected",
    (
        b"Algorithm converged to a point of local "
        b"infeasibility. Problem may be "
        b"infeasible."
    ),
    "Search_Direction_Becomes_Too_Small",
    (b"Algorithm proceeds with very little progress."),
    "Diverging_Iterates",
    b"It seems that the iterates diverge.",
    "User_Requested_Stop",
    (
        b"The user call-back function intermediate_callback "
        b"(see Section 3.3.4 in the documentation) returned "
        b"false, i.e., the user code requested a premature "
        b"termination of the optimization."
    ),
    "Feasible_Point_Found",
    b"Feasible point for square problem found.",
    "Maximum_Iterations_Exceeded",
    (b"Maximum number of iterations exceeded (can be specified by an option)."),
    "Restoration_Failed",
    (b"Restoration phase failed, algorithm doesn't know how to proceed."),
    "Error_In_Step_Computation",
    (
        b"An unrecoverable error occurred while Ipopt "
        b"tried to compute the search direction."
    ),
    "Maximum_CpuTime_Exceeded",
    b"Maximum CPU time exceeded.",
    "Not_Enough_Degrees_Of_Freedom",
    b"Problem has too few degrees of freedom.",
    "Invalid_Problem_Definition",
    b"Invalid problem definition.",
    "Invalid_Option",
    b"Invalid option encountered.",
    "Invalid_Number_Detected",
    (
        b"Algorithm received an invalid number (such as "
        b"NaN or Inf) from the NLP; see also option "
        b"check_derivatives_for_naninf."
    ),
    # Note that the concluding "." was missing before cyipopt 1.0.3
    "Invalid_Number_Detected",
    (
        b"Algorithm received an invalid number (such as "
        b"NaN or Inf) from the NLP; see also option "
        b"check_derivatives_for_naninf"
    ),
    "Unrecoverable_Exception",
    b"Some uncaught Ipopt exception encountered.",
    "NonIpopt_Exception_Thrown",
    b"Unknown Exception caught in Ipopt.",
    # Note that the concluding "." was missing before cyipopt 1.0.3
    "NonIpopt_Exception_Thrown",
    b"Unknown Exception caught in Ipopt",
    "Insufficient_Memory",
    b"Not enough memory.",
    "Internal_Error",
    (
        b"An unknown internal error occurred. Please contact "
        b"the Ipopt authors through the mailing list."
    ),
]
_cyipopt_status_enum = {
    _cyipopt_status_enum[i + 1]: _cyipopt_status_enum[i]
    for i in range(0, len(_cyipopt_status_enum), 2)
}

# This maps Ipopt ApplicationReturnStatus enum strings to an appropriate
# Pyomo TerminationCondition
_ipopt_term_cond = {
    "Solve_Succeeded": TerminationCondition.optimal,
    "Solved_To_Acceptable_Level": TerminationCondition.feasible,
    "Infeasible_Problem_Detected": TerminationCondition.infeasible,
    "Search_Direction_Becomes_Too_Small": TerminationCondition.minStepLength,
    "Diverging_Iterates": TerminationCondition.unbounded,
    "User_Requested_Stop": TerminationCondition.userInterrupt,
    "Feasible_Point_Found": TerminationCondition.feasible,
    "Maximum_Iterations_Exceeded": TerminationCondition.maxIterations,
    "Restoration_Failed": TerminationCondition.noSolution,
    "Error_In_Step_Computation": TerminationCondition.solverFailure,
    "Maximum_CpuTime_Exceeded": TerminationCondition.maxTimeLimit,
    "Not_Enough_Degrees_Of_Freedom": TerminationCondition.invalidProblem,
    "Invalid_Problem_Definition": TerminationCondition.invalidProblem,
    "Invalid_Option": TerminationCondition.error,
    "Invalid_Number_Detected": TerminationCondition.internalSolverError,
    "Unrecoverable_Exception": TerminationCondition.internalSolverError,
    "NonIpopt_Exception_Thrown": TerminationCondition.error,
    "Insufficient_Memory": TerminationCondition.resourceInterrupt,
    "Internal_Error": TerminationCondition.internalSolverError,
}


class CyIpoptSolver:
    def __init__(self, problem_interface, options=None):
        """Create an instance of the CyIpoptSolver. You must
        provide a problem_interface that corresponds to
        the abstract class CyIpoptProblemInterface
        options can be provided as a dictionary of key value
        pairs
        """
        self._problem = problem_interface

        self._options = options
        if options is not None:
            assert isinstance(options, dict)
        else:
            self._options = dict()

    def solve(self, x0=None, tee=False):
        if x0 is None:
            x0 = self._problem.x_init()
        xstart = x0
        cyipopt_solver = self._problem

        # check if we need scaling
        obj_scaling, x_scaling, g_scaling = self._problem.scaling_factors()
        if any(_ is not None for _ in (obj_scaling, x_scaling, g_scaling)):
            # need to set scaling factors
            if obj_scaling is None:
                obj_scaling = 1.0
            if x_scaling is None:
                x_scaling = np.ones(nx)
            if g_scaling is None:
                g_scaling = np.ones(ng)
            try:
                set_scaling = cyipopt_solver.set_problem_scaling
            except AttributeError:
                # Fall back to pre-1.0.0 API
                set_scaling = cyipopt_solver.setProblemScaling
            set_scaling(obj_scaling, x_scaling, g_scaling)

        # add options
        try:
            add_option = cyipopt_solver.add_option
        except AttributeError:
            # Fall back to pre-1.0.0 API
            add_option = cyipopt_solver.addOption
        for k, v in self._options.items():
            add_option(k, v)

        with capture_output(sys.stdout if tee else None, capture_fd=True):
            x, info = cyipopt_solver.solve(xstart)

        return x, info


def _numpy_vector(val):
    ans = np.array(val, np.float64)
    if len(ans.shape) != 1:
        raise ValueError(
            "expected a vector, but received a matrix with shape %s" % (ans.shape,)
        )
    return ans


class PyomoCyIpoptSolver:
    CONFIG = ConfigBlock("cyipopt")
    CONFIG.declare(
        "tee",
        ConfigValue(
            default=False, domain=bool, description="Stream solver output to console"
        ),
    )
    CONFIG.declare(
        "load_solutions",
        ConfigValue(
            default=True,
            domain=bool,
            description="Store the final solution into the original Pyomo model",
        ),
    )
    CONFIG.declare(
        "return_nlp",
        ConfigValue(
            default=False,
            domain=bool,
            description="Return the results object and the underlying nlp"
            " NLP object from the solve call.",
        ),
    )
    CONFIG.declare("options", ConfigBlock(implicit=True))
    CONFIG.declare(
        "intermediate_callback",
        ConfigValue(
            default=None,
            description="Set the function that will be called each iteration.",
        ),
    )
    CONFIG.declare(
        "halt_on_evaluation_error",
        ConfigValue(
            default=None,
            description="Whether to halt if a function or derivative evaluation fails",
        ),
    )

    def __init__(self, **kwds):
        """Create an instance of the CyIpoptSolver. You must
        provide a problem_interface that corresponds to
        the abstract class CyIpoptProblemInterface

        options can be provided as a dictionary of key value
        pairs
        """
        self.config = self.CONFIG(kwds)

    def _set_model(self, model):
        self._model = model

    def available(self, exception_flag=False):
        return bool(numpy_available and cyipopt_interface.cyipopt_available)

    def license_is_valid(self):
        return True

    def version(self):
        def _int(x):
            try:
                return int(x)
            except:
                return x

        return tuple(_int(_) for _ in cyipopt_interface.cyipopt.__version__.split("."))

    def solve(self, model, **kwds):
        config = self.config(kwds, preserve_implicit=True)

        if not isinstance(model, Block):
            raise ValueError(
                "PyomoCyIpoptSolver.solve(model): model must be a Pyomo Block"
            )

        # If this is a Pyomo model / block, then we need to create
        # the appropriate PyomoNLP, then wrap it in a CyIpoptNLP
        grey_box_blocks = list(
            model.component_data_objects(egb.ExternalGreyBoxBlock, active=True)
        )
        # if there is no objective, add one temporarily so we can construct an NLP
        objectives = list(model.component_data_objects(Objective, active=True))
        n_obj = len(objectives)
        for gbb in grey_box_blocks:
            if gbb.get_external_model().has_objective():
                n_obj += 1
        if n_obj == 0:
            objname = unique_component_name(model, "_obj")
            objective = model.add_component(objname, Objective(expr=0.0))
        try:
            if grey_box_blocks:
                # nlp = pyomo_nlp.PyomoGreyBoxNLP(model)
                nlp = pyomo_grey_box.PyomoNLPWithGreyBoxBlocks(model)
            else:
                nlp = pyomo_nlp.PyomoNLP(model)
        finally:
            # We only need the objective to construct the NLP, so we delete
            # it from the model ASAP
            if n_obj == 0:
                model.del_component(objective)

        problem = cyipopt_interface.CyIpoptNLP(
            nlp,
            intermediate_callback=config.intermediate_callback,
            halt_on_evaluation_error=config.halt_on_evaluation_error,
        )
        ng = len(problem.g_lb())
        nx = len(problem.x_lb())
        cyipopt_solver = problem

        # check if we need scaling
        obj_scaling, x_scaling, g_scaling = problem.scaling_factors()
        if any(_ is not None for _ in (obj_scaling, x_scaling, g_scaling)):
            # need to set scaling factors
            if obj_scaling is None:
                obj_scaling = 1.0
            if x_scaling is None:
                x_scaling = np.ones(nx)
            if g_scaling is None:
                g_scaling = np.ones(ng)
            try:
                set_scaling = cyipopt_solver.set_problem_scaling
            except AttributeError:
                # Fall back to pre-1.0.0 API
                set_scaling = cyipopt_solver.setProblemScaling
            set_scaling(obj_scaling, x_scaling, g_scaling)

        # add options
        try:
            add_option = cyipopt_solver.add_option
        except AttributeError:
            # Fall back to pre-1.0.0 API
            add_option = cyipopt_solver.addOption
        for k, v in config.options.items():
            add_option(k, v)

        timer = TicTocTimer()
        try:
            with capture_output(sys.stdout if config.tee else None, capture_fd=True):
                x, info = cyipopt_solver.solve(problem.x_init())
            solverStatus = SolverStatus.ok
        except:
            msg = "Exception encountered during cyipopt solve:"
            logger.error(msg, exc_info=sys.exc_info())
            solverStatus = SolverStatus.unknown
            raise

        wall_time = timer.toc(None)

        results = SolverResults()

        if config.load_solutions:
            nlp.set_primals(x)
            nlp.set_duals(info["mult_g"])
            nlp.load_state_into_pyomo(
                bound_multipliers=(info["mult_x_L"], info["mult_x_U"])
            )
        else:
            soln = Solution()
            sm = nlp.symbol_map
            soln.variable.update(
                (sm.getSymbol(i), {'Value': j, 'ipopt_zL_out': zl, 'ipopt_zU_out': zu})
                for i, j, zl, zu in zip(
                    nlp.get_pyomo_variables(), x, info['mult_x_L'], info['mult_x_U']
                )
            )
            soln.constraint.update(
                (sm.getSymbol(i), {'Dual': j})
                for i, j in zip(nlp.get_pyomo_constraints(), info['mult_g'])
            )
            model.solutions.add_symbol_map(sm)
            results._smap_id = id(sm)
            results.solution.insert(soln)

        results.problem.name = model.name
        obj = next(model.component_data_objects(Objective, active=True))
        results.problem.sense = obj.sense
        if obj.sense == minimize:
            results.problem.upper_bound = info["obj_val"]
        else:
            results.problem.lower_bound = info["obj_val"]
        results.problem.number_of_objectives = 1
        results.problem.number_of_constraints = ng
        results.problem.number_of_variables = nx
        results.problem.number_of_binary_variables = 0
        results.problem.number_of_integer_variables = 0
        results.problem.number_of_continuous_variables = nx
        # TODO: results.problem.number_of_nonzeros

        results.solver.name = "cyipopt"
        results.solver.return_code = info["status"]
        results.solver.message = info["status_msg"]
        results.solver.wallclock_time = wall_time
        status_enum = _cyipopt_status_enum[info["status_msg"]]
        results.solver.termination_condition = _ipopt_term_cond[status_enum]
        results.solver.status = TerminationCondition.to_solver_status(
            results.solver.termination_condition
        )

        problem.close()

        if config.return_nlp:
            return results, nlp

        return results

    #
    # Support "with" statements.
    #
    def __enter__(self):
        return self

    def __exit__(self, t, v, traceback):
        pass
